from typing import Any, Callable, List, Optional, Sequence, Union, Dict

import torch
from llama_index.core.base.llms.types import (
    ChatMessage,
    ChatResponse,
    ChatResponseGen,
    CompletionResponse,
    CompletionResponseGen,
    LLMMetadata,
)

from llama_index.core.bridge.pydantic import Field, PrivateAttr
from llama_index.core.callbacks import CallbackManager
from llama_index.core.constants import (
    DEFAULT_CONTEXT_WINDOW,
    DEFAULT_NUM_OUTPUTS,
)
from llama_index.core.llms.callbacks import (
    llm_chat_callback,
    llm_completion_callback,
)
from llama_index.core.llms.custom import CustomLLM
from llama_index.core.base.llms.generic_utils import (
    completion_response_to_chat_response,
    stream_completion_response_to_chat_response,
    messages_to_prompt as generic_messages_to_prompt,
)
from llama_index.core.prompts.base import PromptTemplate
from llama_index.core.types import BaseOutputParser, PydanticProgramMode, Thread
from transformers import (
    AutoModelForCausalLM,
    AutoTokenizer,
    StoppingCriteria,
    StoppingCriteriaList,
)

from airunner.windows.main.settings_mixin import SettingsMixin
from airunner.settings import AIRUNNER_DEFAULT_LLM_HF_PATH
from airunner.handlers.llm.llm_request import LLMRequest


class HuggingFaceLLM(CustomLLM, SettingsMixin):
    r"""HuggingFace LLM.

    Examples:
        `pip install llama-index-llms-huggingface`

        ```python
        from llama_index.llms.huggingface import HuggingFaceLLM

        def messages_to_prompt(messages):
            prompt = ""
            for message in messages:
                if message.role == 'system':
                prompt += f"<|system|>\n{message.content}</s>\n"
                elif message.role == 'user':
                prompt += f"<|user|>\n{message.content}</s>\n"
                elif message.role == 'assistant':
                prompt += f"<|assistant|>\n{message.content}</s>\n"

            # ensure we start with a system prompt, insert blank if needed
            if not prompt.startswith("<|system|>\n"):
                prompt = "<|system|>\n</s>\n" + prompt

            # add final assistant prompt
            prompt = prompt + "<|assistant|>\n"

            return prompt

        def completion_to_prompt(completion):
            return f"<|system|>\n</s>\n<|user|>\n{completion}</s>\n<|assistant|>\n"

        import torch
        from transformers import BitsAndBytesConfig
        from llama_index.core.prompts import PromptTemplate
        from llama_index.llms.huggingface import HuggingFaceLLM

        # quantize to save memory
        quantization_config = BitsAndBytesConfig(
            load_in_4bit=True,
            bnb_4bit_compute_dtype=torch.float16,
            bnb_4bit_quant_type="nf4",
            bnb_4bit_use_double_quant=True,
        )

        llm = HuggingFaceLLM(
            model_name="HuggingFaceH4/zephyr-7b-beta",
            tokenizer_name="HuggingFaceH4/zephyr-7b-beta",
            context_window=3900,
            max_new_tokens=256,
            model_kwargs={"quantization_config": quantization_config},
            generate_kwargs={"temperature": 0.7, "top_k": 50, "top_p": 0.95},
            messages_to_prompt=messages_to_prompt,
            completion_to_prompt=completion_to_prompt,
            device_map="auto",
        )

        response = llm.complete("What is the meaning of life?")
        print(str(response))
        ```
    """

    model_name: str = Field(
        default=AIRUNNER_DEFAULT_LLM_HF_PATH,
        description=(
            "The model name to use from HuggingFace. "
            "Unused if `model` is passed in directly."
        ),
    )
    model: Optional[Any] = Field(
        default=None,
        description="The model to use. If not passed in, will be loaded from `model_name`.",
    )
    context_window: int = Field(
        default=DEFAULT_CONTEXT_WINDOW,
        description="The maximum number of tokens available for input.",
        gt=0,
    )
    max_new_tokens: int = Field(
        default=DEFAULT_NUM_OUTPUTS,
        description="The maximum number of tokens to generate.",
        gt=0,
    )
    system_prompt: str = Field(
        default="",
        description=(
            "The system prompt, containing any extra instructions or context. "
            "The model card on HuggingFace should specify if this is needed."
        ),
    )
    query_wrapper_prompt: PromptTemplate = Field(
        default=PromptTemplate("{query_str}"),
        description=(
            "The query wrapper prompt, containing the query placeholder. "
            "The model card on HuggingFace should specify if this is needed. "
            "Should contain a `{query_str}` placeholder."
        ),
    )
    tokenizer_name: str = Field(
        default=AIRUNNER_DEFAULT_LLM_HF_PATH,
        description=(
            "The name of the tokenizer to use from HuggingFace. "
            "Unused if `tokenizer` is passed in directly."
        ),
    )
    device_map: str = Field(
        default="auto", description="The device_map to use. Defaults to 'auto'."
    )
    stopping_ids: List[int] = Field(
        default_factory=list,
        description=(
            "The stopping ids to use. "
            "Generation stops when these token IDs are predicted."
        ),
    )
    tokenizer_outputs_to_remove: list = Field(
        default_factory=list,
        description=(
            "The outputs to remove from the tokenizer. "
            "Sometimes huggingface tokenizers return extra inputs that cause errors."
        ),
    )
    tokenizer_kwargs: Dict = Field(
        default_factory=Dict, description="The kwargs to pass to the tokenizer."
    )
    model_kwargs: Dict = Field(
        default_factory=Dict,
        description="The kwargs to pass to the model during initialization.",
    )
    
    _llm_request: Optional[LLMRequest] = None

    @property
    def generate_kwargs(self) -> Dict:
        if self.llm_request:
            data = self.llm_request.to_dict()
            
            # Remove unused keys
            del data["do_tts_reply"]

            return data
        return {}

    @property
    def llm_request(self) -> Optional[LLMRequest]:
        if self._llm_request is None:
            self._llm_request = LLMRequest.from_default()
        return self._llm_request
    
    @llm_request.setter
    def llm_request(self, value: Optional[LLMRequest]):
        self._llm_request = value
        
    is_chat_model: bool = Field(
        default=False,
        description=(
            LLMMetadata.model_fields["is_chat_model"].description
            + " Be sure to verify that you either pass an appropriate tokenizer "
            "that can convert prompts to properly formatted chat messages or a "
            "`messages_to_prompt` that does so."
        ),
    )

    _model: Any = PrivateAttr()
    _tokenizer: Any = PrivateAttr()
    _stopping_criteria: Any = PrivateAttr()
    _streaming_stopping_criteria: Any = PrivateAttr(default=None)

    def __init__(
        self,
        context_window: int = DEFAULT_CONTEXT_WINDOW,
        max_new_tokens: int = DEFAULT_NUM_OUTPUTS,
        query_wrapper_prompt: Union[str, PromptTemplate] = "{query_str}",
        tokenizer_name: str = AIRUNNER_DEFAULT_LLM_HF_PATH,
        model_name: str = AIRUNNER_DEFAULT_LLM_HF_PATH,
        model: Optional[Any] = None,
        tokenizer: Optional[Any] = None,
        device_map: Optional[str] = "auto",
        stopping_ids: Optional[List[int]] = None,
        tokenizer_kwargs: Optional[Dict] = None,
        tokenizer_outputs_to_remove: Optional[list] = None,
        model_kwargs: Optional[Dict] = None,
        generate_kwargs: Optional[Dict] = None,
        is_chat_model: Optional[bool] = False,
        callback_manager: Optional[CallbackManager] = None,
        system_prompt: str = "",
        messages_to_prompt: Optional[Callable[[Sequence[ChatMessage]], str]] = None,
        completion_to_prompt: Optional[Callable[[str], str]] = None,
        pydantic_program_mode: PydanticProgramMode = PydanticProgramMode.DEFAULT,
        output_parser: Optional[BaseOutputParser] = None,
        streaming_stopping_criteria: Optional[StoppingCriteria] = None,
    ) -> None:
        """Initialize params."""
        model_kwargs = model_kwargs or {}
        model = model or AutoModelForCausalLM.from_pretrained(
            model_name, device_map=device_map, **model_kwargs
        )

        # check context_window
        config_dict = model.config.to_dict()
        model_context_window = int(
            config_dict.get("max_position_embeddings", context_window)
        )
        if model_context_window and model_context_window < context_window:
            self.logger.warning(
                f"Supplied context_window {context_window} is greater "
                f"than the model's max input size {model_context_window}. "
                "Disable this warning by setting a lower context_window."
            )
            context_window = model_context_window

        tokenizer_kwargs = tokenizer_kwargs or {}

        tokenizer = tokenizer or AutoTokenizer.from_pretrained(
            tokenizer_name, **tokenizer_kwargs
        )

        # setup stopping criteria
        stopping_ids_list = stopping_ids or []

        class StopOnTokens(StoppingCriteria):
            def __call__(
                self,
                input_ids: torch.LongTensor,
                scores: torch.FloatTensor,
                **kwargs: Any,
            ) -> bool:
                for stop_id in stopping_ids_list:
                    if input_ids[0][-1] == stop_id:
                        return True
                return False

        stopping_criteria = StoppingCriteriaList([StopOnTokens()])

        if isinstance(query_wrapper_prompt, str):
            query_wrapper_prompt = PromptTemplate(query_wrapper_prompt)

        messages_to_prompt = messages_to_prompt or self._tokenizer_messages_to_prompt

        super().__init__(
            context_window=context_window,
            max_new_tokens=max_new_tokens,
            query_wrapper_prompt=query_wrapper_prompt,
            tokenizer_name=tokenizer_name,
            model_name=model_name,
            device_map=device_map,
            stopping_ids=stopping_ids or [],
            tokenizer_kwargs=tokenizer_kwargs or {},
            tokenizer_outputs_to_remove=tokenizer_outputs_to_remove or [],
            model_kwargs=model_kwargs or {},
            generate_kwargs=generate_kwargs or {},
            is_chat_model=is_chat_model,
            callback_manager=callback_manager,
            system_prompt=system_prompt,
            messages_to_prompt=messages_to_prompt,
            completion_to_prompt=completion_to_prompt,
            pydantic_program_mode=pydantic_program_mode,
            output_parser=output_parser,
        )

        self._model = model
        self._tokenizer = tokenizer
        self._stopping_criteria = stopping_criteria
        self._streaming_stopping_criteria = StoppingCriteriaList([streaming_stopping_criteria, StopOnTokens()])
    
    @property
    def model(self):
        return self._model
    
    @property
    def tokenizer(self):
        return self._tokenizer

    @classmethod
    def class_name(cls) -> str:
        return "HuggingFace_LLM"

    @property
    def metadata(self) -> LLMMetadata:
        """LLM metadata."""
        return LLMMetadata(
            context_window=self.context_window,
            num_output=self.max_new_tokens,
            model_name=self.model_name,
            is_chat_model=self.is_chat_model,
            is_function_calling_model=True
        )

    def unload(self):
        del self._model
        del self._tokenizer
        del self._stopping_criteria
        self._model = None
        self._tokenizer = None
        self._stopping_criteria = None

    def _tokenizer_messages_to_prompt(self, messages: Sequence[ChatMessage]) -> str:
        """Use the tokenizer to convert messages to prompt. Fallback to generic."""
        if hasattr(self._tokenizer, "apply_chat_template"):
            messages_dict = [
                {"role": message.role.value, "content": message.content}
                for message in messages
            ]
            return self._tokenizer.apply_chat_template(
                messages_dict, tokenize=False, add_generation_prompt=True
            )

        return generic_messages_to_prompt(messages)

    @llm_completion_callback()
    def complete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponse:
        """Completion endpoint."""
        full_prompt = prompt
        if not formatted:
            if self.query_wrapper_prompt:
                full_prompt = self.query_wrapper_prompt.format(query_str=prompt)
            if self.completion_to_prompt:
                full_prompt = self.completion_to_prompt(full_prompt)
            elif self.system_prompt:
                full_prompt = f"{self.system_prompt} {full_prompt}"

        inputs = self._tokenizer(full_prompt, return_tensors="pt")
        inputs = inputs.to(self._model.device)

        # remove keys from the tokenizer if needed, to avoid HF errors
        for key in self.tokenizer_outputs_to_remove:
            if key in inputs:
                inputs.pop(key, None)

        tokens = self._model.generate(
            **inputs,
            stopping_criteria=self._stopping_criteria,
            **self.generate_kwargs,
        )
        completion_tokens = tokens[0][inputs["input_ids"].size(1):]
        completion = self._tokenizer.decode(completion_tokens, skip_special_tokens=True)

        return CompletionResponse(text=completion, raw={"model_output": tokens})

    @llm_completion_callback()
    def stream_complete(
        self, prompt: str, formatted: bool = False, **kwargs: Any
    ) -> CompletionResponseGen:
        """Streaming completion endpoint."""
        from transformers import TextIteratorStreamer

        full_prompt = prompt
        if not formatted:
            if self.query_wrapper_prompt:
                full_prompt = self.query_wrapper_prompt.format(query_str=prompt)
            if self.system_prompt:
                full_prompt = f"{self.system_prompt} {full_prompt}"

        inputs = self._tokenizer(full_prompt, return_tensors="pt")
        inputs = inputs.to(self._model.device)

        # remove keys from the tokenizer if needed, to avoid HF errors
        for key in self.tokenizer_outputs_to_remove:
            if key in inputs:
                inputs.pop(key, None)

        streamer = TextIteratorStreamer(
            self._tokenizer, skip_prompt=True, skip_special_tokens=True
        )
        generation_kwargs = dict(
            inputs,
            stopping_criteria=self._streaming_stopping_criteria,
            **self.generate_kwargs,
        )
        generation_kwargs["streamer"] = streamer

        # generate in background thread
        # NOTE/TODO: token counting doesn't work with streaming
        thread = Thread(target=self._model.generate, kwargs=generation_kwargs)
        thread.start()

        # create generator based off of streamer
        def gen() -> CompletionResponseGen:
            text = ""
            for x in streamer:
                text += x
                yield CompletionResponse(text=text, delta=x)

        return gen()

    @llm_chat_callback()
    def chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        prompt = self.messages_to_prompt(messages)
        completion_response = self.complete(prompt, formatted=True, **kwargs)
        return completion_response_to_chat_response(completion_response)

    @llm_chat_callback()
    def stream_chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponseGen:
        prompt = self.messages_to_prompt(messages)
        completion_response = self.stream_complete(prompt, formatted=True, **kwargs)
        return stream_completion_response_to_chat_response(completion_response)




from airunner.windows.setup_wizard.model_setup.tts.templates.speecht5_license_ui import Ui_speecht5_license
from airunner.windows.setup_wizard.user_agreement.agreement_page import AgreementPage


class SpeechT5License(AgreementPage):
    class_name_ = Ui_speecht5_license





import random
import os
import torch
from typing import Optional, Dict, List, Union, Type
from peft import PeftModel

from transformers.utils.quantization_config import BitsAndBytesConfig, GPTQConfig
from transformers import AutoModelForCausalLM, AutoTokenizer
from transformers.generation.streamers import TextIteratorStreamer

from llama_index.core.chat_engine.types import AgentChatResponse

from airunner.handlers.base_handler import BaseHandler
from airunner.enums import SignalCode, ModelType, ModelStatus, LLMActionType
from airunner.settings import AIRUNNER_MAX_SEED, AIRUNNER_LOCAL_FILES_ONLY
from airunner.utils.memory import clear_memory
from airunner.handlers.llm.agent.agents import (
    LocalAgent, 
    OpenRouterQObject
)
from airunner.data.models import Conversation, LLMGeneratorSettings
from airunner.handlers.llm.training_mixin import TrainingMixin
from airunner.handlers.llm.llm_request import LLMRequest
from airunner.handlers.llm.llm_response import LLMResponse
from airunner.handlers.llm.llm_settings import LLMSettings


class LLMHandler(
    BaseHandler,
    TrainingMixin
):
    model_type: ModelType = ModelType.LLM
    model_class: str = "llm"
    _model: Optional[AutoModelForCausalLM] = None
    _streamer: Optional[TextIteratorStreamer] = None
    _chat_agent: Optional[Union[Type[LocalAgent], OpenRouterQObject]] = None
    _agent_executor: Optional[object] = None
    _service_context_model: Optional[object] = None
    _use_query_engine: bool = False
    _use_chat_engine: bool = True
    _user_evaluation: str = ""
    _restrict_tools_to_additional: bool = True
    _return_agent_code: bool = False
    _rag_tokenizer: Optional[object] = None
    _rag_retriever: Optional[object] = None
    _vocoder: Optional[object] = None
    _tokenizer: Optional[object] = None
    _generator: Optional[object] = None
    _history: Optional[List] = []
    _current_model_path: Optional[str] = None
    llm_settings: LLMSettings

    def __init__(
        self, 
        local_agent_class: Optional[Type[LocalAgent]], 
        *args, 
        **kwargs
    ):
        self.local_agent_class_ = local_agent_class or LocalAgent
        super().__init__(*args, **kwargs)
        self.llm_settings = LLMSettings()

    @property
    def is_mistral(self) -> bool:
        path = self._current_model_path.lower()
        return "ministral" in path

    @property
    def is_llama_instruct(self):
        path = self._current_model_path.lower()
        if "instruct" in path and "llama" in path:
            return True
        return False

    @property
    def _quantization_config(self):
        config = None
        if self.llm_dtype == "8bit":
            config = BitsAndBytesConfig(
                load_in_4bit=False,
                load_in_8bit=True,
                llm_int8_threshold=6.0,
                llm_int8_has_fp16_weight=False,
                bnb_4bit_compute_dtype=torch.bfloat16,
                bnb_4bit_use_double_quant=True,
                bnb_4bit_quant_type='nf4',
            )
        elif self.llm_dtype == "4bit":
            config = BitsAndBytesConfig(
                load_in_4bit=True,
                bnb_4bit_compute_dtype=torch.float16  # changed to match input type
            )
        elif self.llm_dtype == "2bit":
            config = GPTQConfig(
                bits=2,
                dataset="c4",
                tokenizer=self._tokenizer
            )
        return config

    @property
    def use_cache(self):
        if self.llm_generator_settings.override_parameters:
            return self.llm_generator_settings.use_cache
        return self.chatbot.use_cache

    @property
    def model_version(self) -> str:
        model_version = self.chatbot.model_version
        if self.llm_generator_settings.override_parameters:
            model_version = self.llm_generator_settings.model_version
        return model_version

    @property
    def model_path(self):
        return os.path.expanduser(os.path.join(
            self.path_settings.base_path,
            "text",
            "models",
            "llm",
            "causallm",
            self.model_version
        ))
    
    def load(self):
        if self.model_status in (
            ModelStatus.LOADING,
            ModelStatus.LOADED
        ):
            return
        self.unload()
        self.change_model_status(ModelType.LLM, ModelStatus.LOADING)
        self._current_model_path = self.model_path
        if self.llm_settings.use_local_llm:
            self._load_tokenizer()
            self._load_model()
        self._load_agent()
        if (
            self._model and 
            self._tokenizer and 
            self._chat_agent
        ) or (
            self._chat_agent and
            self.llm_settings.use_api
        ):
            self.change_model_status(ModelType.LLM, ModelStatus.LOADED)
        else:
            if not self._model:
                self.logger.error("Model failed to load")
            if not self._chat_agent:
                self.logger.error("Chat agent failed to load")
            self.change_model_status(ModelType.LLM, ModelStatus.FAILED)

    def unload(self):
        if self.model_status in (
            ModelStatus.LOADING,
            ModelStatus.UNLOADED
        ):
            return
        self.logger.debug("Unloading LLM")
        self.change_model_status(ModelType.LLM, ModelStatus.LOADING)
        self._unload_model()
        self._unload_tokenizer()
        self._unload_agent()
        clear_memory()
        self.change_model_status(ModelType.LLM, ModelStatus.UNLOADED)

    def handle_request(self, data: Dict) -> AgentChatResponse:
        self.logger.debug("Handling request")
        self._do_set_seed()
        self.load()
        return self._do_generate(
            prompt=data["request_data"]["prompt"],
            action=data["request_data"]["action"],
            llm_request=data["request_data"]["llm_request"],
        )

    def do_interrupt(self):
        """
        Public method to interrupt the chat process
        """
        if self._chat_agent:
            self._chat_agent.interrupt_process()

    def on_conversation_deleted(self, data):
        """
        Public method to handle conversation deletion
        """
        if self._chat_agent:
            self._chat_agent.on_conversation_deleted(data)

    def clear_history(self, data: Optional[Dict] = None):
        """
        Public method to clear the chat agent history
        """
        self.logger.debug("Clearing chat history")

        conversation_id = data.get("conversation_id", None)
        llm_generator_settings = LLMGeneratorSettings.objects.first()
        conversation = Conversation.objects.first()
        
        if not conversation_id:
            conversation = Conversation.create()
            data["conversation_id"] = conversation.id
        
        LLMGeneratorSettings.objects.update(
            llm_generator_settings.id,
            current_conversation=conversation.id
        )

        if self._chat_agent:
            self._chat_agent.clear_history(data)

    def add_chatbot_response_to_history(self, message):
        """
        Public method to add a chatbot response to the chat agent history
        """
        self._chat_agent.add_chatbot_response_to_history(message)

    def load_conversation(self, message):
        """
        Public method to load a conversation into the chat agent
        """
        self._chat_agent.on_load_conversation(message)

    def reload_rag_engine(self):
        """
        Public method to reload the RAG model
        """
        self._chat_agent.reload_rag_engine()

    def on_section_changed(self):
        self._chat_agent.current_tab = None
    
    def on_web_browser_page_html(self, content: str):
        if self._chat_agent:
            self._chat_agent.on_web_browser_page_html(content)
        else:
            self.logger.error("Chat agent not loaded")

    def _load_tokenizer(self):
        if self._tokenizer is not None:
            return
        self.logger.debug(f"Loading tokenizer from {self.model_path}")
        try:
            self._tokenizer = AutoTokenizer.from_pretrained(
                self.model_path,
                local_files_only=AIRUNNER_LOCAL_FILES_ONLY,
                device_map=self.device,
                trust_remote_code=False,
                torch_dtype=self.torch_dtype,
                attn_implementation="flash_attention_2",
            )
            self.logger.debug("Tokenizer loaded")
        except Exception as e:
            self.logger.error(e)

        if self._tokenizer:
            self._tokenizer.use_default_system_prompt = False
        else:
            self.logger.error("Tokenizer failed to load")

    def _load_model(self):
        if self._model is not None:
            return        
        self.logger.debug(f"Loading local LLM model from {self.model_path}")
        try:
            # Use the same path as tokenizer
            self._model = AutoModelForCausalLM.from_pretrained(
                self.model_path,
                local_files_only=AIRUNNER_LOCAL_FILES_ONLY,
                use_cache=self.use_cache,
                trust_remote_code=False,
                torch_dtype=self.torch_dtype,
                device_map=self.device,
            )
        except Exception as e:
            self.logger.error(f"Error loading model: {e}")
            return
        
        try:
            if os.path.exists(self.adapter_path):
                # Convert base model to PEFT format
                self._model = PeftModel.from_pretrained(
                    self._model, 
                    self.adapter_path
                )
        except Exception as e:
            self.logger.error(
                f"Error loading adapter (continuing with base model): {e}"
            )

    def _load_agent(self):
        if self._chat_agent is not None:
            return
        self.logger.debug("Loading agent")
        if self.llm_settings.use_local_llm:
            self.logger.info("Loading local chat agent")
            self._chat_agent = self.local_agent_class_(
                model=self._model,
                tokenizer=self._tokenizer,
                default_tool_choice=None,
                llm_settings=self.llm_settings
            )
        elif self.llm_settings.use_openrouter:
            self.logger.info("Loading openrouter chat agent")
            self._chat_agent = OpenRouterQObject(
                llm_settings=self.llm_settings
            )
        else:
            self.logger.warning("No chat agent to load")
        self.logger.info("Chat agent loaded")

    def _unload_model(self):
        self.logger.debug("Unloading model")
        self._model = None
        return True

    def _unload_tokenizer(self):
        self.logger.debug("Unloading tokenizer")
        try:
            del self._tokenizer
        except AttributeError as e:
            self.logger.warning(f"Error unloading tokenizer {e}")
        self._tokenizer = None
        clear_memory(self.memory_settings.default_gpu_llm)
        return True

    def _unload_agent(self):
        self.logger.debug("Unloading agent")
        do_clear_memory = False
        if self._chat_agent is not None:
            self.logger.debug("Unloading chat agent")
            self._chat_agent.unload()
            try:
                del self._chat_agent
            except AttributeError as e:
                self.logger.warning(f"Error unloading chat agent: {e}")
            self._chat_agent = None
            do_clear_memory = True
        return do_clear_memory
        
    def _do_generate(
        self, 
        prompt: str, 
        action: LLMActionType,
        system_prompt: Optional[str] = None,
        rag_system_prompt: Optional[str] = None,
        llm_request: Optional[LLMRequest] = None,
        do_tts_reply: bool = True
    ) -> AgentChatResponse:
        self.logger.debug("Generating response")
        if self._current_model_path != self.model_path:
            self.unload()
            self.load()
        # if action is LLMActionType.CHAT and self.chatbot.use_mood:
        #     action = LLMActionType.UPDATE_MOOD
        response = self._chat_agent.chat(
            message=prompt, 
            action=action, 
            system_prompt=system_prompt, 
            rag_system_prompt=rag_system_prompt,
            llm_request=llm_request
        )
        if action is LLMActionType.CHAT:
            self._send_final_message()
        return response

    def _send_final_message(self):
        self.logger.debug("Sending final message")
        self.logger.debug("Emitting streamed text signal")
        self.emit_signal(SignalCode.LLM_TEXT_STREAMED_SIGNAL, {
            "response": LLMResponse(is_end_of_message=True)
        })

    def _do_set_seed(self):
        self.logger.debug("Setting seed")

        if self.llm_generator_settings.override_parameters:
            seed = self.llm_generator_settings.seed
            random_seed = self.llm_generator_settings.random_seed
        else:
            seed = self.chatbot.seed
            random_seed = self.chatbot.random_seed

        if random_seed:
            seed = random.randint(-AIRUNNER_MAX_SEED, AIRUNNER_MAX_SEED)

        torch.manual_seed(seed)
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        random.seed(seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False
        if self._tokenizer:
            self._tokenizer.seed = seed



from typing import Any
from llama_index.core.tools.types import ToolOutput
from llama_index.core.base.llms.types import ChatMessage, MessageRole

from airunner.handlers.llm.agent.chat_engine import ReactAgentEngine
from airunner.handlers.llm.agent.tools.chat_engine_tool import ChatEngineTool


class ReActAgentTool(ChatEngineTool):
    """ReActAgentTool.
    
    A tool for determining which actions to take.
    """
    @classmethod
    def from_tools(cls, *args, **kwargs) -> "ReActAgentTool":
        agent = kwargs.pop("agent", None)
        return_direct = kwargs.pop("return_direct", False)
        chat_engine = ReactAgentEngine.from_tools(*args, **kwargs)
        name = "react_agent_tool"
        description = """Useful for determining which tool to use."""
        return cls.from_defaults(
            chat_engine=chat_engine, 
            name=name,
            description=description,
            return_direct=return_direct,
            agent=agent
        )
    
    def call(self, *args: Any, **kwargs: Any) -> ToolOutput:
        query_str = self._get_query_str(*args, **kwargs)
        chat_history = kwargs.get("chat_history", None)
        tool_choice = kwargs.get("tool_choice", None)
        streaming_response = self.chat_engine.stream_chat(
            query_str, 
            chat_history=chat_history if chat_history else [],
            tool_choice=tool_choice
        )
        self.chat_engine.chat_history.append(
            ChatMessage(
                content=query_str,
                role=MessageRole.USER
            )
        )

        response = ""
        is_first_message = True
        for token in streaming_response.response_gen:
            response += token
            self.agent.handle_response(token, is_first_message)
            is_first_message = False
        
        self.chat_engine.chat_history.append(
            ChatMessage(
                content=response,
                role=MessageRole.ASSISTANT
            )
        )

        return ToolOutput(
            content=str(response),
            tool_name=self.metadata.name,
            raw_input={"input": query_str},
            raw_output=response,
        )



from abc import ABCMeta, ABC

from bs4 import BeautifulSoup
from llama_index.core.readers.base import BasePydanticReader
from llama_index.core.schema import Document

BasePydanticReaderMeta = type(BasePydanticReader)


class CombinedMeta(BasePydanticReaderMeta, ABCMeta):
    pass


class HtmlFileReader(BasePydanticReader, ABC, metaclass=CombinedMeta):
    @staticmethod
    def read(file_path):
        with open(file_path, 'r', encoding='utf-8') as file:
            content = file.read()
        soup = BeautifulSoup(content, 'html.parser')
        return soup.get_text()

    def load_data(self, *args, **load_kwargs):
        documents = []
        file_path = load_kwargs["extra_info"]["file_path"]
        documents.append(Document(
            text=self.read(file_path),
            id_=load_kwargs["extra_info"]["file_name"],
            metadata=load_kwargs["extra_info"])
        )
        return documents



import os
import threading

from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.image.image_widget import ImageWidget
from airunner.widgets.image.templates.image_panel_widget_ui import Ui_image_panel_widget
from airunner.widgets.qflowlayout.q_flow_layout import QFlowLayout


class ImagePanelWidget(BaseWidget):
    widget_class_ = Ui_image_panel_widget

    def __init__(self, *args, **kwargs):
        """
        Initializes the ImagePanelWidget.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            None
        """
        super().__init__(*args, **kwargs)
        self.page = 0
        self.total_per_page = 50
        self.page_step = 512
        self.last_page = False
        self.sorted_files = []
        self.start = 0
        self.end = 0
        self.ui.scrollArea.verticalScrollBar().valueChanged.connect(self.handle_scroll)
        flow_layout = QFlowLayout()
        self.ui.scrollAreaWidgetContents.setLayout(flow_layout)
        self.display_thread = threading.Thread(target=self.display_thumbnails)

    def showEvent(self, event):
        super().showEvent(event)
        if self.path_settings.image_path != "":
            self.load_files()
            self.show_files()
        else:
            self.logger.warning("Image path not set. Please set the image path in the settings.")

    def clear_files(self):
        self.page = 0
        # remove all images from the container
        while self.ui.scrollAreaWidgetContents.layout().count() > 0:
            item = self.ui.scrollAreaWidgetContents.layout().takeAt(0)
            widget = item.widget()
            widget.deleteLater()

    def show_files(self, reset_scroll_bar=True):
        """
        Displays the files in the image panel widget.

        Args:
            reset_scroll_bar (bool, optional): Whether to reset the scroll bar position. Defaults to True.
        """
        if reset_scroll_bar:
            self.ui.scrollArea.verticalScrollBar().setValue(0)

        self.start = self.page * self.total_per_page
        self.end = self.start + self.total_per_page

        self.display_thumbnails()

    def load_files(self):
        """
        Load files from the specified image path and sort them based on modification time.

        Returns:
            None
        """
        image_path = self.path_settings.image_path
        files_in_image_path = os.listdir(image_path)
        sorted_files = {}
        for file in files_in_image_path:
            if os.path.isdir(os.path.join(image_path, file)):
                sorted_files[file] = []
                for root, dirs, files_in_dir in os.walk(os.path.join(image_path, file)):
                    files = []
                    for f in files_in_dir:
                        if ".png.thumbnail.png" not in f:
                            files.append(os.path.join(root, f))
                    sorted_files[file].extend(files)
        
        section = "txt2img"
        files = sorted_files[section]
        files.sort(key=os.path.getmtime, reverse=True)
        self.sorted_files = files

    def display_thumbnails(self):
        """
        Display thumbnails of images from the sorted_files list within the specified range.
        """
        for file in self.sorted_files[self.start:self.end]:
            if file.endswith(".png"):
                image_widget = ImageWidget(self, is_thumbnail=True)
                image_widget.set_image(os.path.join(self.path_settings.image_path, file))
                self.ui.scrollAreaWidgetContents.layout().addWidget(image_widget)
    
    def handle_scroll(self, value):
        """
        Handles the scroll event of the image panel widget.

        Args:
            value (int): The value of the scroll event.

        Returns:
            None
        """
        if value >= self.ui.scrollArea.verticalScrollBar().maximum() - self.page_step + 1:
            self.page += 1
            self.show_files(reset_scroll_bar=False)




from PySide6.QtCore import Slot
from PySide6.QtWidgets import QInputDialog, QMessageBox

from airunner.data.models import TargetFiles, Chatbot
from airunner.enums import SignalCode
from airunner.utils import open_file_path
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.llm.document_widget import DocumentWidget
from airunner.widgets.llm.templates.bot_preferences_ui import Ui_bot_preferences


class BotPreferencesWidget(BaseWidget):
    widget_class_ = Ui_bot_preferences

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def showEvent(self, event):
        self.load_saved_chatbots()
        super().showEvent(event)

    def load_form_elements(self):
        elements = [
            "botname",
            "bot_personality",
            "names_groupbox",
            "personality_groupbox",
            "system_instructions",
            "system_instructions_groupbox",
            "guardrails_prompt",
            "guardrails_groupbox",
            "target_files",
            "use_weather_prompt",
            "use_datetime",
        ]
        self.toggle_signals(self.ui, elements)
        self.ui.botname.setText(self.chatbot.botname)
        self.ui.bot_personality.setPlainText(self.chatbot.bot_personality)
        self.ui.names_groupbox.setChecked(self.chatbot.assign_names)
        self.ui.personality_groupbox.setChecked(self.chatbot.use_personality)
        self.ui.system_instructions.setPlainText(self.chatbot.system_instructions)
        self.ui.system_instructions_groupbox.setChecked(self.chatbot.use_system_instructions)
        self.ui.guardrails_prompt.setPlainText(self.chatbot.guardrails_prompt)
        self.ui.guardrails_groupbox.setChecked(self.chatbot.use_guardrails)
        self.ui.use_weather_prompt.setChecked(self.chatbot.use_weather_prompt)
        self.ui.use_datetime.setChecked(self.chatbot.use_datetime)
        self.load_documents()
        self.toggle_signals(self.ui, elements, False)

    @staticmethod
    def toggle_signals(ui: object, elements: list, block: bool = True):
        for element in elements:
            getattr(ui, element).blockSignals(block)

    def botname_changed(self, val):
        self.update_chatbot("botname", val)

    def bot_personality_changed(self):
        self.update_chatbot("bot_personality", self.ui.bot_personality.toPlainText())

    def guardrails_prompt_changed(self):
        self.update_chatbot("guardrails_prompt", self.ui.guardrails_prompt.toPlainText())

    def system_instructions_changed(self):
        self.update_chatbot("system_instructions", self.ui.system_instructions.toPlainText())

    def toggle_use_names(self, val):
        self.update_chatbot("assign_names", val)

    def toggle_use_personality(self, val):
        self.update_chatbot("use_personality", val)

    def toggle_use_guardrails(self, val):
        self.update_chatbot("use_guardrails", val)

    def toggle_use_system_instructions(self, val):
        self.update_chatbot("use_system_instructions", val)

    def create_new_chatbot_clicked(self):
        # Display a dialog asking for the name of the new chatbot
        chatbot_name, ok = QInputDialog.getText(self, "New Chatbot", "Enter the name of the new chatbot:")

        # If the user clicked "OK" and entered a name
        if ok and chatbot_name:
            self.create_chatbot(chatbot_name)
            self.update_llm_generator_settings("current_chatbot", chatbot_name)
            self.load_saved_chatbots()

    @Slot(str)
    def saved_chatbots_changed(self, val: str):
        chatbot = Chatbot.objects.filter_first(Chatbot.name == val)
        chatbot_id = chatbot.id
        self.update_llm_generator_settings("current_chatbot", chatbot_id)
        self.load_form_elements()
        self.emit_signal(SignalCode.CHATBOT_CHANGED)

    def load_saved_chatbots(self):
        names = [chatbot.name for chatbot in self.chatbots]
        self.ui.saved_chatbots.blockSignals(True)
        self.ui.saved_chatbots.clear()
        self.ui.saved_chatbots.addItems(names)
        self.ui.saved_chatbots.setCurrentIndex(self.ui.saved_chatbots.findText(self.chatbot.name))
        self.load_form_elements()
        self.ui.saved_chatbots.blockSignals(False)

    def delete_clicked(self):
        # show confirmation before deleting
        if self.chatbot.name == "Default":
            return
        msg = f"Are you sure you want to delete {self.chatbot.name}?"
        reply = self.show_confirmation_dialog(msg)
        if reply == 0:
            self.delete_chatbot_by_name(self.chatbot.name)
            self.chatbot.name = "Chatbot"
            self.load_saved_chatbots()

    def show_confirmation_dialog(self, msg):
        return self.show_dialog(msg, "Confirmation", buttons=["Yes", "No"])

    @staticmethod
    def show_dialog(msg, title, buttons=["OK"]):
        dialog = QMessageBox()
        dialog.setText(msg)
        dialog.setWindowTitle(title)
        for button in buttons:
            dialog.addButton(button, QMessageBox.ButtonRole.AcceptRole)
        return dialog.exec()

    @Slot(bool)
    def toggle_use_image_generator(self, val: bool):
        self.ui.use_image_generator_checkbox.blockSignals(True)
        self.ui.use_image_generator_checkbox.setChecked(val)
        self.ui.use_image_generator_checkbox.blockSignals(False)

    @Slot(bool)
    def agent_type_changed(self, val: str):
        print("agent type changed", val)

    @Slot(bool)
    def use_weather_prompt_toggled(self, val: bool):
        self.update_chatbot("use_weather_prompt", val)

    @Slot(bool)
    def toggle_use_datetime(self, val: bool):
        self.update_chatbot("use_datetime", val)

    @Slot()
    def browse_documents(self):
        file_path = open_file_path(self, file_type="Text Files (*.md *.html *.htm *.epub *.pdf *.txt)")

        # validate file path
        if not file_path or not file_path[0] or not file_path[0].strip() or not file_path[0].endswith((
            ".md",
            ".html",
            ".htm",
            ".epub",
            ".pdf",
            ".txt",
        )):
            self.logger.error(f"Invalid file path: {file_path}")
            return

        self.add_chatbot_document_to_chatbot(self.chatbot, file_path[0])
        self.emit_signal(SignalCode.RAG_RELOAD_INDEX_SIGNAL)
        self.load_documents()

    def load_documents(self):
        # Get the layout of the scroll area's widget
        layout = self.ui.target_files.widget().layout()

        # Clear the existing widgets in the layout
        for i in reversed(range(layout.count())):
            layout.itemAt(i).widget().setParent(None)

        for target_file in self.chatbot.target_files:
            widget = DocumentWidget(target_file, self.delete_document)
            layout.addWidget(widget)

    def delete_document(self, target_file: TargetFiles):
        TargetFiles.objects.delete(target_file.id)
        
        self.load_documents()
        self.emit_signal(SignalCode.RAG_RELOAD_INDEX_SIGNAL)

    def update_chatbot(self, key, val):
        chatbot = self.chatbot
        try:
            setattr(chatbot, key, val)
        except TypeError:
            self.logger.error(f"Attribute {key} does not exist in Chatbot")
            return
        chatbot.save()



from PySide6.QtCore import QObject, QPropertyAnimation, QEasingCurve, Property, QPointF, Slot, QSize, Signal
from PySide6.QtGui import QGradient, QLinearGradient, QPalette, Qt, QPainter, QColor
from PySide6.QtWidgets import QAbstractButton, QApplication


class SwitchPrivate(QObject):
    def __init__(self, q, parent=None):
        QObject.__init__(self, parent=parent)
        self.mPointer = q
        self.mPosition = 0.0
        self.mGradient = QLinearGradient()
        self.mGradient.setSpread(QGradient.Spread.PadSpread)
        self.checked = False

        self.animation = QPropertyAnimation(self)
        self.animation.setTargetObject(self)
        self.animation.setPropertyName(b'position')
        self.animation.setStartValue(0.0)
        self.animation.setEndValue(1.0)
        self.animation.setDuration(200)
        self.animation.setEasingCurve(QEasingCurve.Type.InOutExpo)

        self.animation.finished.connect(self.mPointer.update)

    @Property(float)
    def position(self):
        return self.mPosition

    @position.setter
    def position(self, value):
        self.mPosition = value
        self.mPointer.update()

    def draw(self, painter):
        r = self.mPointer.rect()
        margin = r.height() / 10
        shadow = self.mPointer.palette().color(QPalette.ColorRole.Shadow)
        light = self.mPointer.palette().color(QPalette.ColorRole.Light)
        button = self.mPointer.palette().color(QPalette.ColorRole.Button)
        painter.setPen(Qt.PenStyle.NoPen)

        if self.mPointer.isEnabled() and self.checked:
            self.mGradient.setColorAt(0, Qt.GlobalColor.blue)
            self.mGradient.setColorAt(1, Qt.GlobalColor.blue)
        else:
            self.mGradient.setColorAt(0, shadow.darker(130))
            self.mGradient.setColorAt(1, light.darker(130))

        self.mGradient.setStart(0, r.height())
        self.mGradient.setFinalStop(0, 0)
        painter.setBrush(self.mGradient)
        painter.drawRoundedRect(r, r.height() / 2, r.height() / 2)

        self.mGradient.setColorAt(0, shadow.darker(140))
        self.mGradient.setColorAt(1, light.darker(160))
        self.mGradient.setStart(0, 0)
        self.mGradient.setFinalStop(0, r.height())
        painter.setBrush(self.mGradient)
        painter.drawRoundedRect(r.adjusted(margin, margin, -margin, -margin), r.height() / 2, r.height() / 2)

        self.mGradient.setColorAt(0, button.darker(130))
        self.mGradient.setColorAt(1, button)

        painter.setBrush(self.mGradient)

        x = r.height() / 2.0 + self.mPosition * (r.width() - r.height())
        painter.drawEllipse(QPointF(x, r.height() / 2), r.height() / 2 - margin, r.height() / 2 - margin)

    @Slot(bool, name='animate')
    def animate(self, checked):
        self.checked = checked
        self.animation.setDirection(
            QPropertyAnimation.Direction.Forward if checked else QPropertyAnimation.Direction.Backward
        )
        self.animation.start()


class SwitchWidget(QAbstractButton):
    toggled = Signal(bool)

    def __init__(self, parent=None):
        QAbstractButton.__init__(self, parent=parent)
        self.dPtr = SwitchPrivate(self)
        self.setCheckable(True)
        self.clicked.connect(self.dPtr.animate)
        self.clicked.connect(self.emitToggled)
        self._backgroundColor = QColor("blue")  # Initialize the internal attribute
        self.setChecked(False)
        self.dPtr.animate(False)

    def emitToggled(self, checked):
        self.toggled.emit(checked)

    def sizeHint(self):
        return QSize(84, 42)

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        self.dPtr.draw(painter)

    def resizeEvent(self, event):
        self.update()

    @Property(QColor)
    def backgroundColor(self):
        return self._backgroundColor  # Return the internal attribute

    @backgroundColor.setter
    def backgroundColor(self, color):
        self._backgroundColor = color  # Set the internal attribute
        self.update()


if __name__ == '__main__':
    import sys
    app = QApplication(sys.argv)
    w = SwitchWidget()
    w.setProperty("backgroundColor", QColor("blue"))
    w.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'llm_history_item.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QFrame, QGridLayout, QHBoxLayout,
    QLabel, QPushButton, QSizePolicy, QSpacerItem,
    QTextBrowser, QWidget)

class Ui_llm_history_item_widget(object):
    def setupUi(self, llm_history_item_widget):
        if not llm_history_item_widget.objectName():
            llm_history_item_widget.setObjectName(u"llm_history_item_widget")
        llm_history_item_widget.resize(384, 217)
        self.gridLayout = QGridLayout(llm_history_item_widget)
        self.gridLayout.setSpacing(10)
        self.gridLayout.setObjectName(u"gridLayout")
        self.gridLayout.setContentsMargins(10, 0, 0, 0)
        self.conversation_description = QTextBrowser(llm_history_item_widget)
        self.conversation_description.setObjectName(u"conversation_description")

        self.gridLayout.addWidget(self.conversation_description, 0, 0, 1, 1)

        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setSpacing(5)
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.horizontalSpacer = QSpacerItem(40, 20, QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Minimum)

        self.horizontalLayout.addItem(self.horizontalSpacer)

        self.timestamp = QLabel(llm_history_item_widget)
        self.timestamp.setObjectName(u"timestamp")

        self.horizontalLayout.addWidget(self.timestamp)

        self.line = QFrame(llm_history_item_widget)
        self.line.setObjectName(u"line")
        self.line.setFrameShape(QFrame.Shape.VLine)
        self.line.setFrameShadow(QFrame.Shadow.Sunken)

        self.horizontalLayout.addWidget(self.line)

        self.botname = QLabel(llm_history_item_widget)
        self.botname.setObjectName(u"botname")

        self.horizontalLayout.addWidget(self.botname)

        self.line_2 = QFrame(llm_history_item_widget)
        self.line_2.setObjectName(u"line_2")
        self.line_2.setFrameShape(QFrame.Shape.VLine)
        self.line_2.setFrameShadow(QFrame.Shadow.Sunken)

        self.horizontalLayout.addWidget(self.line_2)

        self.pushButton = QPushButton(llm_history_item_widget)
        self.pushButton.setObjectName(u"pushButton")
        icon = QIcon(QIcon.fromTheme(u"document-open"))
        self.pushButton.setIcon(icon)

        self.horizontalLayout.addWidget(self.pushButton)

        self.pushButton_2 = QPushButton(llm_history_item_widget)
        self.pushButton_2.setObjectName(u"pushButton_2")
        icon1 = QIcon(QIcon.fromTheme(u"user-trash"))
        self.pushButton_2.setIcon(icon1)

        self.horizontalLayout.addWidget(self.pushButton_2)


        self.gridLayout.addLayout(self.horizontalLayout, 1, 0, 1, 1)


        self.retranslateUi(llm_history_item_widget)
        self.pushButton.clicked.connect(llm_history_item_widget.action_load_conversation_clicked)
        self.pushButton_2.clicked.connect(llm_history_item_widget.action_delete_conversation_clicked)

        QMetaObject.connectSlotsByName(llm_history_item_widget)
    # setupUi

    def retranslateUi(self, llm_history_item_widget):
        llm_history_item_widget.setWindowTitle(QCoreApplication.translate("llm_history_item_widget", u"Form", None))
        self.timestamp.setText(QCoreApplication.translate("llm_history_item_widget", u"Date", None))
        self.botname.setText(QCoreApplication.translate("llm_history_item_widget", u"Bot name", None))
#if QT_CONFIG(tooltip)
        self.pushButton.setToolTip(QCoreApplication.translate("llm_history_item_widget", u"Load conversation", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton.setText("")
#if QT_CONFIG(tooltip)
        self.pushButton_2.setToolTip(QCoreApplication.translate("llm_history_item_widget", u"Deleted conversation", None))
#endif // QT_CONFIG(tooltip)
        self.pushButton_2.setText("")
    # retranslateUi




import os
import json

from PySide6.QtGui import QPixmap
from PySide6.QtWidgets import QLabel
from PySide6.QtCore import Qt
from PySide6.QtCore import Signal
from PySide6.QtWidgets import QMenu
from PySide6.QtWidgets import QMessageBox
from PIL import Image
from PIL.ImageQt import ImageQt
from airunner.enums import SignalCode, CanvasToolName
from airunner.utils.image import load_metadata_from_image, delete_image
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.image.templates.image_widget_ui import Ui_image_widget
from PySide6.QtGui import QDrag
from PySide6.QtCore import QMimeData
from PySide6.QtCore import QByteArray


class ImageWidget(BaseWidget):
    widget_class_ = Ui_image_widget
    image_path = None
    meta_data = {}
    image_width = 0
    image_height = 0
    clicked = Signal()
    pixmap = None

    def __init__(self, *args, **kwargs):
        self.is_thumbnail = kwargs.pop("is_thumbnail", False)
        super().__init__(*args, **kwargs)
    
    def set_image(self, image_path):
        size = self.ui.image_frame.width()
        self.image_path = image_path

        self.load_meta_data(image_path)

        # Create a QPixmap object

        if isinstance(self.image_path, Image.Image):
            qimage = ImageQt(self.image_path)  # Convert the PngImageFile to a QImage
            pixmap = QPixmap.fromImage(qimage)  # Create a QPixmap from the QImage
        else:
            # check if thumbnail exists at path
            # if not, create it
            path = self.image_path + ".thumbnail.png"
            if not os.path.exists(path):
                image = Image.open(self.image_path)
                try:
                    image.thumbnail((size, size))
                except OSError:
                    pass
                try:
                    image.save(path)
                except OSError:
                    pass
            if self.is_thumbnail:
                pixmap = QPixmap(path)
            else:
                pixmap = QPixmap(self.image_path)
        self.pixmap = pixmap

        # set width and height
        self.image_width = pixmap.width()
        self.image_height = pixmap.height()

        # Create a QLabel object
        label = QLabel(self.ui.image_frame)

        # set width and height of label to size
        label.setFixedWidth(size)
        label.setFixedHeight(size)
        label.mousePressEvent = self.handle_label_clicked
        label.setCursor(Qt.CursorShape.PointingHandCursor)

        # Set the pixmap to the label
        label.setPixmap(pixmap)
        label.setAlignment(Qt.AlignmentFlag.AlignCenter)
    
    def handle_label_clicked(self, event):
        if event.button() == Qt.MouseButton.LeftButton:
            drag = QDrag(self)
            mime_data = QMimeData()

            # Load the image metadata
            self.load_meta_data(self.image_path)

            # Convert the metadata to a JSON string
            meta_data = self.meta_data.copy()
            meta_data["path"] = self.image_path
            meta_data_str = json.dumps(meta_data)

            # Encode the JSON string to bytes
            meta_data_bytes = meta_data_str.encode()

            # Set the metadata as additional data in the QMimeData object
            mime_data.setData("application/x-qt-image-metadata", QByteArray(meta_data_bytes))

            drag.setMimeData(mime_data)

            # Set a pixmap for the drag operation
            pixmap = QPixmap(self.image_path)

            # Scale the pixmap to no larger than 128x128 while maintaining aspect ratio
            pixmap = pixmap.scaled(128, 128, Qt.AspectRatioMode.KeepAspectRatio)

            drag.setPixmap(pixmap)

            # Execute the drag operation
            drag.exec(Qt.DropAction.MoveAction)
        elif event.button() == Qt.MouseButton.RightButton:
            # self.send_image_to_grid()
            self.display_image_menu(event)

    def display_image_menu(self, event):
        context_menu = QMenu(self)

        view_action = context_menu.addAction("View")
        edit_action = context_menu.addAction("Edit")
        delete_action = context_menu.addAction("Delete")
        
        # display image in a window
        view_action.triggered.connect(lambda: self.view_image())
        edit_action.triggered.connect(lambda: self.send_image_to_grid())
        delete_action.triggered.connect(lambda: self.delete_image())

        global_position = self.mapToGlobal(event.pos())
        context_menu.exec(global_position)
    
    def load_meta_data(self, image_path):
        # load the png metadata from image_path
        # check if image_path is Image
        if isinstance(image_path, Image.Image):
            image = image_path
        else:
            with open(image_path, 'rb') as image_file:
                image = Image.open(image_file)
        
        if image:
            self.meta_data = load_metadata_from_image(image)

    def send_image_to_grid(self):
        self.emit_signal(SignalCode.CANVAS_LOAD_IMAGE_FROM_PATH_SIGNAL, {
            "image_path": self.image_path
        })

    def view_image(self):
        from PySide6.QtWidgets import QGraphicsView, QGraphicsScene, QDialog, QVBoxLayout
        from PySide6.QtGui import QPixmap
        from PySide6.QtCore import Qt
        from PySide6.QtGui import QPainter
        # Open the image
        image = QPixmap(self.image_path)

        # Create a QGraphicsScene and add the image to it
        scene = QGraphicsScene()
        scene.addPixmap(image)

        # Create a QGraphicsView and set its scene to be the one we just created
        view = QGraphicsView(scene)

        # Enable dragging and using the scroll wheel to zoom
        view.setRenderHint(QPainter.RenderHint.Antialiasing)
        view.setRenderHint(QPainter.RenderHint.SmoothPixmapTransform)
        view.setRenderHint(QPainter.RenderHint.TextAntialiasing)
        view.setOptimizationFlags(
            QGraphicsView.OptimizationFlag.DontAdjustForAntialiasing |
            QGraphicsView.OptimizationFlag.DontSavePainterState
        )
        view.setViewportUpdateMode(QGraphicsView.ViewportUpdateMode.FullViewportUpdate)
        view.setTransformationAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)
        view.setResizeAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)
        view.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        view.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        view.setInteractive(True)
        view.setMouseTracking(True)
        view.setRenderHints(
            QPainter.RenderHint.Antialiasing |
            QPainter.RenderHint.SmoothPixmapTransform |
            QPainter.RenderHint.TextAntialiasing
        )

        # Create a layout and add the QGraphicsView to it
        layout = QVBoxLayout()
        layout.addWidget(view)

        # Create a QDialog, set its layout and show it
        dialog = QDialog()
        dialog.setLayout(layout)
        dialog.exec()

    def delete_image(self):
        confirm_dialog = QMessageBox()
        confirm_dialog.setIcon(QMessageBox.Icon.Warning)
        confirm_dialog.setText("Are you sure you want to delete this image?")
        confirm_dialog.setWindowTitle("Confirm Delete")
        confirm_dialog.setStandardButtons(QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No)
        
        result = confirm_dialog.exec()
        if result == QMessageBox.StandardButton.Yes:
            self.confirm_delete()
        else:
            self.cancel_delete()

    def confirm_delete(self):
        delete_image(self.image_path)
        delete_image(self.image_path + ".thumbnail.png")
        self.deleteLater()
    
    def cancel_delete(self):
        pass


class BrushImageWidget(ImageWidget):
    def __init__(self, *args, **kwargs):
        self.container = kwargs.pop("container", None)
        self.brush = kwargs.pop("brush", CanvasToolName.BRUSH)
        super().__init__(*args, **kwargs)

    def handle_label_clicked(self, event):
        # get the clicked object
        if event.button() == Qt.MouseButton.LeftButton:
            shift_pressed = event.modifiers() == Qt.KeyboardModifier.ShiftModifier
            self.emit_signal(SignalCode.PRESET_IMAGE_GENERATOR_ACTIVATE_BRUSH_SIGNAL, {
                'event': event,
                'widget': self,
                'brush': self.brush,
                'shift_pressed': shift_pressed
            })
        elif event.button() == Qt.MouseButton.RightButton:
            self.emit_signal(SignalCode.PRESET_IMAGE_GENERATOR_DISPLAY_ITEM_MENU_SIGNAL, {
                'event': event,
                'widget': self,
                'brush': self.brush
            })



import numpy as np
from PIL import Image, ImageChops

from airunner.filters.base_filter import BaseFilter


class RGBNoiseFilter(BaseFilter):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.red_grain = None
        self.green_grain = None
        self.blue_grain = None

    def apply_filter(self, image, do_reset):
        # Convert image to numpy array
        image_array = np.array(image)

        # Generate random noise
        red_noise = np.random.rand(*image.size) * self.red
        green_noise = np.random.rand(*image.size) * self.green
        blue_noise = np.random.rand(*image.size) * self.blue

        # Add noise to each channel
        image_array[..., 0] = np.clip(image_array[..., 0] + red_noise, 0, 255)
        image_array[..., 1] = np.clip(image_array[..., 1] + green_noise, 0, 255)
        image_array[..., 2] = np.clip(image_array[..., 2] + blue_noise, 0, 255)

        # Convert back to PIL image
        image = Image.fromarray(image_array.astype('uint8'), 'RGBA')
        return image



from abc import abstractmethod, ABCMeta, ABC
from typing import Optional, Type, ClassVar

from transformers import PreTrainedModel, ProcessorMixin

from airunner.handlers.base_handler import BaseHandler
from airunner.enums import ModelType
from airunner.utils import prepare_text_for_tts


BaseHandlerMeta = type(BaseHandler)


class CombinedMeta(BaseHandlerMeta, ABCMeta):
    pass


class TTSHandler(BaseHandler, ABC, metaclass=CombinedMeta):
    """
    Abstract base class for text-to-speech handlers.
    Responsible for managing the model, processor, vocoder, and speaker embeddings.

    Use from a worker to avoid blocking the main thread.
    """
    # These should remain class attributes as they define class-wide properties
    # that identify which model/processor classes to use for this handler type
    target_model: ClassVar[Optional[str]] = None
    model_class: ClassVar[Optional[Type[PreTrainedModel]]] = None
    processor_class: ClassVar[Optional[Type[ProcessorMixin]]] = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.model_type = ModelType.TTS
        self._engine = None
        
        # Runtime instances of the models should be instance attributes
        self._model = None
        self._processor = None
    
    @abstractmethod
    def reload_speaker_embeddings(self):
        """
        Reload speaker embeddings.
        """

    @abstractmethod
    def interrupt_process_signal(self):
        """
        Signal to interrupt the current TTS process.
        """
    
    @abstractmethod
    def offload_to_cpu(self):
        """
        Move model to CPU to free up GPU memory.
        """
    
    @abstractmethod
    def move_to_device(self, device=None):
        """
        Move model to the specified device.
        """

    @abstractmethod
    def generate(self, message):
        """
        Generate speech from text.
        """

    @staticmethod
    def _prepare_text(text: str) -> str:
        """Prepare text for TTS processing using utility functions."""
        return prepare_text_for_tts(text)



from airunner.handlers.base_handler import BaseHandler
from airunner.handlers.llm.llm_handler import LLMHandler
from airunner.handlers.tts.tts_handler import TTSHandler
from airunner.handlers.tts.speecht5_handler import SpeechT5Handler
from airunner.handlers.tts.openvoice_handler import OpenVoiceHandler
from airunner.handlers.tts.espeak_handler import EspeakHandler
from airunner.handlers.stt.whisper_handler import WhisperHandler
from airunner.handlers.stablediffusion.stablediffusion_handler import StableDiffusionHandler


__all__ = [
    "BaseHandler",
    "LLMHandler",
    "TTSHandler",
    "SpeechT5Handler",
    "OpenVoiceHandler",
    "EspeakHandler",
    "WhisperHandler",
    "StableDiffusionHandler",
]




from airunner.enums import SignalCode
from airunner.utils import random_seed
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.seed.templates.seed_ui import Ui_seed_widget


class SeedWidget(BaseWidget):
    widget_class_ = Ui_seed_widget
    name = "seed_widget"

    def __init__(self, *args, **kwargs):
        self.signal_handlers = {
            SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL: self.on_application_settings_changed_signal
        }
        self.icons = [
            ("dice-game-icon", "random_button")
        ]
        super().__init__(*args, **kwargs)
        self.ui.lineEdit.blockSignals(True)
        self.ui.lineEdit.setText(str(self.generator_settings.seed))
        self.ui.lineEdit.setEnabled(not self.generator_settings.random_seed)
        self.ui.random_button.setChecked(self.generator_settings.random_seed)
        self.ui.lineEdit.blockSignals(False)

    def on_application_settings_changed_signal(self):
        try:
            self.ui.lineEdit.blockSignals(True)
            self.ui.lineEdit.setText(str(self.generator_settings.seed))
            self.ui.lineEdit.blockSignals(False)
        except RuntimeError as _e:
            pass

    def action_clicked_button_random_seed(self, value):
        self.update_generator_settings("random_seed", value)
        self.ui.lineEdit.setEnabled(not value)
        if value is True:
            seed = random_seed()
            self.update_generator_settings("seed", seed)
            self.ui.lineEdit.setText(str(seed))

    def action_value_changed_seed(self, value):
        self.update_generator_settings("seed", int(value))



from typing import Optional
from alembic import op
import sqlalchemy as sa


def get_tables():
    conn = op.get_bind()
    inspector = sa.inspect(conn)
    return inspector.get_table_names()


def table_exists(table_name):
    conn = op.get_bind()
    inspector = sa.inspect(conn)
    return table_name in inspector.get_table_names()


def add_table(cls):
    if not table_exists(cls.__tablename__):
        columns = [column.copy() for column in cls.__table__.columns]
        op.create_table(cls.__tablename__, *columns, *getattr(cls, '__table_args__', ()))
    else:
        print(f"Table '{cls.__tablename__}' already exists, skipping add.")
    return


def add_tables(classes):
    for cls in classes:
        create_table_with_defaults(cls)
    return


def drop_table(cls: Optional[object] = None, table_name: Optional[str] = None):
    if cls is not None:
        table_name = cls.__tablename__
    if table_exists(table_name):
        op.drop_table(table_name)
    else:
        print(f"Table '{table_name}' does not exist, skipping drop.")
    return


def drop_tables(classes):
    for cls in classes:
        drop_table(cls)
    return


def create_table_with_defaults(model):
    if not table_exists(model.__tablename__):
        try:
            columns = []
            for column in model.__table__.columns:
                column_copy = column.copy()
                if column.default is not None:
                    column_copy.server_default = column.default
                columns.append(column_copy)
            op.create_table(
                model.__tablename__,
                *columns,
                *getattr(model, '__table_args__', ())
            )
            set_default_values(model)
        except Exception as e:
            print(f"Failed to create table {model.__tablename__}: {str(e)}")
    else:
        print(f"{model.__tablename__} already exists, skipping")


def set_default_values(model):
    default_values = {}
    for column in model.__table__.columns:
        if column.default is not None:
            default_values[column.name] = column.default.arg
    op.bulk_insert(
        model.__table__,
        [default_values]
    )



import importlib.util
import os
import sys


class PluginLoader:
    def __init__(self, plugin_dir):
        self.plugin_dir = plugin_dir

    def load_plugins(self):
        plugins = []
        if not os.path.exists(self.plugin_dir):
            os.makedirs(self.plugin_dir)

        for foldername in os.listdir(self.plugin_dir):
            plugin_path = os.path.join(self.plugin_dir, foldername, "plugin.py")
            path = os.path.join(self.plugin_dir, foldername)

            # Append the plugin directory to sys.path temporarily
            if os.path.isfile(plugin_path):
                original_sys_path = sys.path.copy()
                sys.path.append(path)

                try:
                    module_name = f"plugin_{foldername}"  # Use a unique module name
                    spec = importlib.util.spec_from_file_location(module_name, plugin_path)
                    module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(module)

                    # Check if the module has a `Plugin` class
                    if hasattr(module, "Plugin"):
                        plugins.append(module.Plugin())
                finally:
                    # Restore the original sys.path to avoid conflicts
                    sys.path = original_sys_path

        return plugins



import os

from airunner.data.models import PathSettings


def create_airunner_paths(path_settings: PathSettings):
    """
    This function creates directories based on the paths provided in the path_settings dictionary.

    The function validates the paths, sanitizes them by removing "..", checks if the application has
    the necessary permissions to create a directory in the specified path, and handles any exceptions
    that might occur when creating the directory.

    Parameters:
    path_settings (dict): A dictionary where the keys are the names of the paths and the values are
    the paths themselves.

    Returns:
    None
    """
    print("Creating directories...")
    for attr in (
        "base_path",
        "documents_path",
        "ebook_path",
        "image_path",
        "llama_index_path",
        "webpages_path",
        "stt_model_path",
        "tts_model_path",
    ):
        path = getattr(path_settings, attr)
        # Path sanitization
        path = path.replace('..', '')

        # Permission check and directory creation
        try:
            path = os.path.expanduser(path)
            if not os.path.exists(path):
                os.makedirs(path)
        except PermissionError:
            print(f"No permission to create directory at {path}")
        except Exception as e:
            print(f"Failed to create directory {path}: {e}")



from typing import Callable
from PySide6.QtCore import QObject, QThread, Signal
from airunner.handlers.stablediffusion import DownloadWorker
from airunner.enums import SignalCode
from airunner.mediator_mixin import MediatorMixin
from airunner.windows.main.settings_mixin import SettingsMixin


class HuggingfaceDownloader(
    MediatorMixin,
    SettingsMixin,
    QObject,
):
    completed = Signal()

    def __init__(self, callback=None):
        super().__init__()
        self.thread = None
        self.worker = None
        self.downloading = False

        self.thread = QThread()
        self.worker = DownloadWorker()
        self.worker.moveToThread(self.thread)

        # Connect signals
        self.worker.finished.connect(lambda: self.emit_signal(SignalCode.DOWNLOAD_COMPLETE))
        self.worker.progress.connect(lambda current, total: callback(current, total))

        self.logger.debug(f"Starting model download thread")
        self.thread.finished.connect(self.handle_completed)
        self.thread.started.connect(self.worker.download)
        self.thread.start()

    def download_model(
        self,
        requested_path: str,
        requested_file_name: str,
        requested_file_path: str,
        requested_callback: Callable[[int, int], None]
    ):
        self.worker.add_to_queue((
            requested_path,
            requested_file_name,
            requested_file_path,
            requested_callback
        ))

    def handle_completed(self):
        self.completed.emit()
        self.worker.deleteLater()
        self.thread.deleteLater()

    def stop_download(self):
        if self.worker:
            self.worker.cancel()
            self.remove_file()



################################################################
# Importing this module sets the Hugging Face environment
# variables for the application.
################################################################
from typing import Optional, Dict
import os.path
import sys
import signal
import traceback
from functools import partial
from pathlib import Path
from importlib.metadata import version

from PySide6 import QtCore
from PySide6.QtCore import (
    QObject,
    QTimer
)
from PySide6.QtGui import (
    QGuiApplication,
    QPixmap,
    Qt,
    QWindow
)
from PySide6.QtWidgets import (
    QApplication,
    QSplashScreen
)

from airunner.app_installer import AppInstaller
from airunner.enums import SignalCode
from airunner.mediator_mixin import MediatorMixin
from airunner.windows.main.settings_mixin import SettingsMixin
from airunner.data.models import ApplicationSettings
from airunner.windows.main.main_window import MainWindow
from airunner.settings import AIRUNNER_DISCORD_URL, AIRUNNER_DISABLE_SETUP_WIZARD


class App(
    MediatorMixin,
    SettingsMixin,
    QObject
):
    """
    The main application class for AI Runner.
    This class can be run as a GUI application or as a socket server.
    """
    def __init__(
        self,
        no_splash: bool = False,
        main_window_class: QWindow = None,
        window_class_params: Optional[Dict] = None
    ):
        """
        Initialize the application and run as a GUI application or a socket server.
        :param main_window_class: The main window class to use for the application.
        """
        self.main_window_class_ = main_window_class or MainWindow
        self.window_class_params = window_class_params or {}
        self.no_splash = no_splash
        self.app = None
        self.splash = None

        """
        Mediator and Settings mixins are initialized here, enabling the application
        to easily access the application settings dictionary.
        """
        super().__init__()

        self.register(SignalCode.LOG_LOGGED_SIGNAL, self.on_log_logged_signal)

        self.start()
        self.run_setup_wizard()

        current_version = version("airunner")
        if self.do_upgrade(current_version):
            self.handle_upgrade(current_version)
        self.run()
    
    def do_upgrade(self, current_version) -> bool:
        current_version_tuple = tuple(map(int, current_version.split(".")))
        try:
            app_version_tuple = tuple(map(int, self.application_settings.app_version.split(".")))
        except ValueError:
            return True
        except AttributeError:
            return True
        return app_version_tuple < current_version_tuple

    def handle_upgrade(self, current_version):
        from airunner.data.bootstrap.pipeline_bootstrap_data import pipeline_bootstrap_data
        from airunner.data.models import PipelineModel
        for model in pipeline_bootstrap_data:
            pipelinemodel = PipelineModel.objects.filter_by_first(
                pipeline_action=model["pipeline_action"],
                version=model["version"],
                category=model["category"],
                classname=model["classname"],
                default=model["default"]
            )
            if pipelinemodel:
                continue
            pipelinemodel = PipelineModel()
            pipelinemodel.pipeline_action = model["pipeline_action"]
            pipelinemodel.version = model["version"]
            pipelinemodel.category = model["category"]
            pipelinemodel.classname = model["classname"]
            pipelinemodel.default = model["default"]
            pipelinemodel.save()
        try:
            app_version_tuple = tuple(map(int, self.application_settings.app_version.split(".")))
        except ValueError:
            app_version_tuple = None
        except AttributeError:
            app_version_tuple = None
        if app_version_tuple is None or app_version_tuple < (3, 1, 10):
            turbo_paths = (
                os.path.expanduser(os.path.join(
                    self.path_settings.base_path, "art/models", "SDXL 1.0", "txt2img", "turbo_models"
                )),
                os.path.expanduser(os.path.join(
                    self.path_settings.base_path, "art/models", "SDXL 1.0", "inpaint", "turbo_models"
                ))
            )
            for turbo_path in turbo_paths:
                if not os.path.exists(turbo_path):
                    os.makedirs(turbo_path)
                    with open(os.path.join(turbo_path, "README.txt"), "w") as f:
                        f.write("Place Stable Diffusion XL Turbo, Lightning and Hyper models here")
            self.application_settings.run_setup_wizard = True
            self.run_setup_wizard()
            self.application_settings.app_version = current_version
            self.llm_generator_settings.model_version = "w4ffl35/Ministral-8B-Instruct-2410-doublequant"
            self.application_settings.save()
            self.llm_generator_settings.save()

    @staticmethod
    def run_setup_wizard():
        if AIRUNNER_DISABLE_SETUP_WIZARD:
            return
        application_settings = ApplicationSettings.objects.first()
        if application_settings.run_setup_wizard:
            AppInstaller()

    def on_log_logged_signal(self, data: dict):
        message = data["message"].split(" - ")
        self.update_splash_message(self.splash, message[4])

    def start(self):
        """
        Conditionally initialize and display the setup wizard.
        :return:
        """
        signal.signal(signal.SIGINT, self.signal_handler)
        QApplication.setAttribute(Qt.ApplicationAttribute.AA_UseDesktopOpenGL)
        self.app = QApplication([])

    def run(self):
        """
        Run as a GUI application.
        A splash screen is displayed while the application is loading
        and a main window is displayed once the application is ready.

        Override this method to run the application in a different mode.
        """
        # Continue with application execution
        if not self.no_splash:
            self.splash = self.display_splash_screen(self.app)

        # Show the main application window
        QTimer.singleShot(
            50,
            partial(
                self.show_main_application,
                self.app
            )
        )
        sys.exit(self.app.exec())

    @staticmethod
    def signal_handler(
        _signal,
        _frame
    ):
        """
        Handle the SIGINT signal in a clean way.
        :param _signal:
        :param _frame:
        :return: None
        """
        print("\nExiting...")
        try:
            app = QApplication.instance()
            app.quit()
            sys.exit(0)
        except Exception as e:
            print(e)
            sys.exit(0)

    def display_splash_screen(self, app):
        """
        Display a splash screen while the application is loading.
        :param app:
        :return:
        """
        if self.no_splash:
            return
        
        screens = QGuiApplication.screens()
        try:
            screen = screens.at(0)
        except AttributeError:
            screen = screens[0]

        base_dir = Path(os.path.dirname(os.path.realpath(__file__)))
        stylesheet_path = base_dir / "images" / "splashscreen.png"
        pixmap = QPixmap(stylesheet_path)
        splash = QSplashScreen(
            screen,
            pixmap,
            QtCore.Qt.WindowType.WindowStaysOnTopHint
        )
        splash.show()
        App.update_splash_message(splash, f"Loading AI Runner")
        app.processEvents()
        return splash

    @staticmethod
    def update_splash_message(splash, message: str):
        splash.showMessage(
            message,
            QtCore.Qt.AlignmentFlag.AlignBottom | QtCore.Qt.AlignmentFlag.AlignCenter,
            QtCore.Qt.GlobalColor.white
        )

    def show_main_application(
        self,
        app
    ):
        """
        Show the main application window.
        :param app:
        :param splash:
        :return:
        """
        try:
            window = self.main_window_class_(
                app=self,
                **self.window_class_params
            )
        except Exception as e:
            traceback.print_exc()
            print(e)
            if self.splash:
                self.splash.finish(None)
            sys.exit(f"""
                An error occurred while initializing the application.
                Please report this issue on GitHub or Discord {AIRUNNER_DISCORD_URL}."
            """)
        app.main_window = window
        if self.splash:
            self.splash.finish(window)
        window.raise_()



from PySide6.QtCore import QPoint
from PySide6.QtGui import QPainter
from PySide6.QtWidgets import QGraphicsItem, QGraphicsPixmapItem

from airunner.enums import CanvasToolName
from airunner.mediator_mixin import MediatorMixin
from airunner.utils import snap_to_grid
from airunner.windows.main.settings_mixin import SettingsMixin


class DraggablePixmap(
    MediatorMixin,
    SettingsMixin,
    QGraphicsPixmapItem,
):
    def __init__(self, pixmap):
        super().__init__(pixmap)
        self.pixmap = pixmap
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable, True)
        self.last_pos = QPoint(0, 0)
        self.save = False
        self.setFlag(
            QGraphicsItem.GraphicsItemFlag.ItemIsMovable,
            True
        )
        x = self.drawing_pad_settings.x_pos
        y = self.drawing_pad_settings.y_pos
        self.setPos(QPoint(x, y))

    @property
    def current_tool(self):
        return CanvasToolName(self.application_settings.current_tool)

    def mouseMoveEvent(self, event):
        if self.current_tool not in [
            CanvasToolName.ACTIVE_GRID_AREA,
            CanvasToolName.SELECTION
        ]:
            return
        super().mouseMoveEvent(event)
        self.snap_to_grid()

    def mouseReleaseEvent(self, event):
        if self.current_tool in [
            CanvasToolName.ACTIVE_GRID_AREA,
            CanvasToolName.SELECTION
        ]:
            self.snap_to_grid(save=True)
        super().mouseReleaseEvent(event)

    def snap_to_grid(self, x=None, y=None, save=False):
        if x is None:
            x = int(self.x())
        if y is None:
            y = int(self.y())
        if self.grid_settings.snap_to_grid:
            x, y = snap_to_grid(self.grid_settings, x, y, False)
        x += self.last_pos.x()
        y += self.last_pos.y()
        self.update_position(x, y, save)

    def update_position(self, x: int, y: int, save: bool = True):
        self.setPos(QPoint(x, y))
        if save:
            self.update_drawing_pad_settings("x_pos", x)
            self.update_drawing_pad_settings("y_pos", y)

    def paint(self, painter: QPainter, option, widget=None):
        painter.drawPixmap(self.pixmap.rect(), self.pixmap)

    def setPixmap(self, pixmap):
        self.pixmap = pixmap
        super().setPixmap(pixmap)



from airunner.filters.base_filter import BaseFilter
from PIL import ImageOps


class Invert(BaseFilter):
    def apply_filter(self, image, do_reset):
        # Ensure the image is in RGB mode
        image = image.convert("RGB")
        # Invert the colors
        return ImageOps.invert(image)



import os
import subprocess


def show_path(path):
    path = os.path.expanduser(path)
    if not os.path.isdir(path):
        print("not a path", path)
        return

    from sys import platform
    platform = platform.lower()
    if platform in ["windows", "win32", "cygwin"]:
        subprocess.Popen(["explorer", os.path.realpath(path)])
    elif platform in ["linux", "linux2"]:
        subprocess.Popen(["xdg-open", os.path.realpath(path)])
    else:
        print("unsupported platform", platform)



from typing import Optional
from dataclasses import dataclass

from airunner.enums import LLMActionType


@dataclass
class LLMResponse:
    message: str = ""
    is_first_message: bool = False
    is_end_of_message: bool = False
    name: Optional[str] = None
    action: LLMActionType = LLMActionType.CHAT



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'flan_generator_form.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QCheckBox, QComboBox, QGridLayout,
    QGroupBox, QHBoxLayout, QLineEdit, QSizePolicy,
    QWidget)

from airunner.widgets.slider.slider_widget import SliderWidget

class Ui_flan_generator_form(object):
    def setupUi(self, flan_generator_form):
        if not flan_generator_form.objectName():
            flan_generator_form.setObjectName(u"flan_generator_form")
        flan_generator_form.resize(400, 631)
        self.gridLayout = QGridLayout(flan_generator_form)
        self.gridLayout.setObjectName(u"gridLayout")
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.groupBox = QGroupBox(flan_generator_form)
        self.groupBox.setObjectName(u"groupBox")
        self.gridLayout_2 = QGridLayout(self.groupBox)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.top_p = SliderWidget(self.groupBox)
        self.top_p.setObjectName(u"top_p")
        self.top_p.setProperty("min_slider", 1)
        self.top_p.setProperty("max_slider", 100)
        self.top_p.setProperty("min_spinbox", 0.000000000000000)
        self.top_p.setProperty("max_spinbox", 1.000000000000000)
        self.top_p.setProperty("display_as_float", True)
        self.top_p.setProperty("slider_step", 1)
        self.top_p.setProperty("slider_page", 10)
        self.top_p.setProperty("spinbox_step", 0.010000000000000)
        self.top_p.setProperty("spinbox_page", 0.100000000000000)

        self.gridLayout_2.addWidget(self.top_p, 0, 0, 1, 1)


        self.horizontalLayout_2.addWidget(self.groupBox)

        self.groupBox_2 = QGroupBox(flan_generator_form)
        self.groupBox_2.setObjectName(u"groupBox_2")
        self.gridLayout_3 = QGridLayout(self.groupBox_2)
        self.gridLayout_3.setObjectName(u"gridLayout_3")
        self.max_length = SliderWidget(self.groupBox_2)
        self.max_length.setObjectName(u"max_length")
        self.max_length.setProperty("min_slider", 1)
        self.max_length.setProperty("max_slider", 2556)
        self.max_length.setProperty("min_spinbox", 1.000000000000000)
        self.max_length.setProperty("max_spinbox", 2556.000000000000000)
        self.max_length.setProperty("display_as_float", False)
        self.max_length.setProperty("slider_step", 1)
        self.max_length.setProperty("slider_page", 2556)
        self.max_length.setProperty("spinbox_step", 1)
        self.max_length.setProperty("spinbox_page", 2556)

        self.gridLayout_3.addWidget(self.max_length, 0, 0, 1, 1)


        self.horizontalLayout_2.addWidget(self.groupBox_2)


        self.gridLayout.addLayout(self.horizontalLayout_2, 1, 0, 1, 1)

        self.random_seed = QCheckBox(flan_generator_form)
        self.random_seed.setObjectName(u"random_seed")

        self.gridLayout.addWidget(self.random_seed, 10, 0, 1, 1)

        self.groupBox_11 = QGroupBox(flan_generator_form)
        self.groupBox_11.setObjectName(u"groupBox_11")
        self.gridLayout_4 = QGridLayout(self.groupBox_11)
        self.gridLayout_4.setObjectName(u"gridLayout_4")
        self.seed = QLineEdit(self.groupBox_11)
        self.seed.setObjectName(u"seed")

        self.gridLayout_4.addWidget(self.seed, 0, 0, 1, 1)


        self.gridLayout.addWidget(self.groupBox_11, 9, 0, 1, 1)

        self.horizontalLayout_4 = QHBoxLayout()
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.groupBox_5 = QGroupBox(flan_generator_form)
        self.groupBox_5.setObjectName(u"groupBox_5")
        self.gridLayout_7 = QGridLayout(self.groupBox_5)
        self.gridLayout_7.setObjectName(u"gridLayout_7")
        self.length_penalty = SliderWidget(self.groupBox_5)
        self.length_penalty.setObjectName(u"length_penalty")
        self.length_penalty.setProperty("min_slider", 0)
        self.length_penalty.setProperty("max_slider", 200)
        self.length_penalty.setProperty("min_spinbox", 0.000000000000000)
        self.length_penalty.setProperty("max_spinbox", 2.000000000000000)
        self.length_penalty.setProperty("display_as_float", True)
        self.length_penalty.setProperty("slider_step", 1)
        self.length_penalty.setProperty("slider_page", 10)
        self.length_penalty.setProperty("spinbox_step", 0.010000000000000)
        self.length_penalty.setProperty("spinbox_page", 0.100000000000000)

        self.gridLayout_7.addWidget(self.length_penalty, 0, 0, 1, 1)


        self.horizontalLayout_4.addWidget(self.groupBox_5)

        self.groupBox_6 = QGroupBox(flan_generator_form)
        self.groupBox_6.setObjectName(u"groupBox_6")
        self.gridLayout_8 = QGridLayout(self.groupBox_6)
        self.gridLayout_8.setObjectName(u"gridLayout_8")
        self.num_beams = SliderWidget(self.groupBox_6)
        self.num_beams.setObjectName(u"num_beams")
        self.num_beams.setProperty("min_slider", 1)
        self.num_beams.setProperty("max_slider", 100)
        self.num_beams.setProperty("min_spinbox", 0.000000000000000)
        self.num_beams.setProperty("max_spinbox", 100.000000000000000)
        self.num_beams.setProperty("display_as_float", False)
        self.num_beams.setProperty("slider_step", 1)
        self.num_beams.setProperty("slider_page", 10)
        self.num_beams.setProperty("spinbox_step", 0.010000000000000)
        self.num_beams.setProperty("spinbox_page", 0.100000000000000)

        self.gridLayout_8.addWidget(self.num_beams, 0, 0, 1, 1)


        self.horizontalLayout_4.addWidget(self.groupBox_6)


        self.gridLayout.addLayout(self.horizontalLayout_4, 3, 0, 1, 1)

        self.do_sample = QCheckBox(flan_generator_form)
        self.do_sample.setObjectName(u"do_sample")

        self.gridLayout.addWidget(self.do_sample, 7, 0, 1, 1)

        self.model_version = QComboBox(flan_generator_form)
        self.model_version.setObjectName(u"model_version")

        self.gridLayout.addWidget(self.model_version, 0, 0, 1, 1)

        self.early_stopping = QCheckBox(flan_generator_form)
        self.early_stopping.setObjectName(u"early_stopping")

        self.gridLayout.addWidget(self.early_stopping, 8, 0, 1, 1)

        self.horizontalLayout_3 = QHBoxLayout()
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.groupBox_3 = QGroupBox(flan_generator_form)
        self.groupBox_3.setObjectName(u"groupBox_3")
        self.gridLayout_5 = QGridLayout(self.groupBox_3)
        self.gridLayout_5.setObjectName(u"gridLayout_5")
        self.repetition_penalty = SliderWidget(self.groupBox_3)
        self.repetition_penalty.setObjectName(u"repetition_penalty")
        self.repetition_penalty.setProperty("min_slider", 0)
        self.repetition_penalty.setProperty("max_slider", 10000)
        self.repetition_penalty.setProperty("min_spinbox", 0.000000000000000)
        self.repetition_penalty.setProperty("max_spinbox", 100.000000000000000)
        self.repetition_penalty.setProperty("display_as_float", True)
        self.repetition_penalty.setProperty("slider_step", 0)
        self.repetition_penalty.setProperty("slider_page", 1)
        self.repetition_penalty.setProperty("spinbox_step", 1.000000000000000)
        self.repetition_penalty.setProperty("spinbox_page", 10.000000000000000)

        self.gridLayout_5.addWidget(self.repetition_penalty, 0, 0, 1, 1)


        self.horizontalLayout_3.addWidget(self.groupBox_3)

        self.groupBox_4 = QGroupBox(flan_generator_form)
        self.groupBox_4.setObjectName(u"groupBox_4")
        self.gridLayout_6 = QGridLayout(self.groupBox_4)
        self.gridLayout_6.setObjectName(u"gridLayout_6")
        self.min_length = SliderWidget(self.groupBox_4)
        self.min_length.setObjectName(u"min_length")
        self.min_length.setProperty("min_slider", 1)
        self.min_length.setProperty("max_slider", 2556)
        self.min_length.setProperty("min_spinbox", 1.000000000000000)
        self.min_length.setProperty("max_spinbox", 2556.000000000000000)
        self.min_length.setProperty("display_as_float", False)
        self.min_length.setProperty("slider_step", 1)
        self.min_length.setProperty("slider_page", 2556)
        self.min_length.setProperty("spinbox_step", 1)
        self.min_length.setProperty("spinbox_page", 2556)

        self.gridLayout_6.addWidget(self.min_length, 0, 0, 1, 1)


        self.horizontalLayout_3.addWidget(self.groupBox_4)


        self.gridLayout.addLayout(self.horizontalLayout_3, 2, 0, 1, 1)

        self.horizontalLayout_5 = QHBoxLayout()
        self.horizontalLayout_5.setObjectName(u"horizontalLayout_5")
        self.groupBox_7 = QGroupBox(flan_generator_form)
        self.groupBox_7.setObjectName(u"groupBox_7")
        self.gridLayout_9 = QGridLayout(self.groupBox_7)
        self.gridLayout_9.setObjectName(u"gridLayout_9")
        self.ngram_size = SliderWidget(self.groupBox_7)
        self.ngram_size.setObjectName(u"ngram_size")
        self.ngram_size.setProperty("min_slider", 1)
        self.ngram_size.setProperty("max_slider", 100)
        self.ngram_size.setProperty("min_spinbox", 0.000000000000000)
        self.ngram_size.setProperty("max_spinbox", 1.000000000000000)
        self.ngram_size.setProperty("display_as_float", True)
        self.ngram_size.setProperty("slider_step", 1)
        self.ngram_size.setProperty("slider_page", 10)
        self.ngram_size.setProperty("spinbox_step", 0.010000000000000)
        self.ngram_size.setProperty("spinbox_page", 0.100000000000000)

        self.gridLayout_9.addWidget(self.ngram_size, 0, 0, 1, 1)


        self.horizontalLayout_5.addWidget(self.groupBox_7)

        self.groupBox_8 = QGroupBox(flan_generator_form)
        self.groupBox_8.setObjectName(u"groupBox_8")
        self.gridLayout_10 = QGridLayout(self.groupBox_8)
        self.gridLayout_10.setObjectName(u"gridLayout_10")
        self.temperature = SliderWidget(self.groupBox_8)
        self.temperature.setObjectName(u"temperature")
        self.temperature.setProperty("min_slider", 1)
        self.temperature.setProperty("max_slider", 200)
        self.temperature.setProperty("min_spinbox", 0.000000000000000)
        self.temperature.setProperty("max_spinbox", 2.000000000000000)
        self.temperature.setProperty("display_as_float", True)
        self.temperature.setProperty("slider_step", 1)
        self.temperature.setProperty("slider_page", 10)
        self.temperature.setProperty("spinbox_step", 0.010000000000000)
        self.temperature.setProperty("spinbox_page", 0.100000000000000)

        self.gridLayout_10.addWidget(self.temperature, 0, 0, 1, 1)


        self.horizontalLayout_5.addWidget(self.groupBox_8)


        self.gridLayout.addLayout(self.horizontalLayout_5, 4, 0, 1, 1)

        self.horizontalLayout_6 = QHBoxLayout()
        self.horizontalLayout_6.setObjectName(u"horizontalLayout_6")
        self.groupBox_9 = QGroupBox(flan_generator_form)
        self.groupBox_9.setObjectName(u"groupBox_9")
        self.gridLayout_11 = QGridLayout(self.groupBox_9)
        self.gridLayout_11.setObjectName(u"gridLayout_11")
        self.sequences = SliderWidget(self.groupBox_9)
        self.sequences.setObjectName(u"sequences")
        self.sequences.setProperty("min_slider", 1)
        self.sequences.setProperty("max_slider", 100)
        self.sequences.setProperty("min_spinbox", 0.000000000000000)
        self.sequences.setProperty("max_spinbox", 100.000000000000000)
        self.sequences.setProperty("display_as_float", False)
        self.sequences.setProperty("slider_step", 1)
        self.sequences.setProperty("slider_page", 10)
        self.sequences.setProperty("spinbox_step", 0.010000000000000)
        self.sequences.setProperty("spinbox_page", 0.100000000000000)

        self.gridLayout_11.addWidget(self.sequences, 0, 0, 1, 1)


        self.horizontalLayout_6.addWidget(self.groupBox_9)

        self.groupBox_10 = QGroupBox(flan_generator_form)
        self.groupBox_10.setObjectName(u"groupBox_10")
        self.gridLayout_12 = QGridLayout(self.groupBox_10)
        self.gridLayout_12.setObjectName(u"gridLayout_12")
        self.top_k = SliderWidget(self.groupBox_10)
        self.top_k.setObjectName(u"top_k")
        self.top_k.setProperty("min_slider", 1)
        self.top_k.setProperty("max_slider", 256)
        self.top_k.setProperty("min_spinbox", 0.000000000000000)
        self.top_k.setProperty("max_spinbox", 256.000000000000000)
        self.top_k.setProperty("display_as_float", False)
        self.top_k.setProperty("slider_step", 1)
        self.top_k.setProperty("slider_page", 10)
        self.top_k.setProperty("spinbox_step", 1)
        self.top_k.setProperty("spinbox_page", 10)

        self.gridLayout_12.addWidget(self.top_k, 0, 0, 1, 1)


        self.horizontalLayout_6.addWidget(self.groupBox_10)


        self.gridLayout.addLayout(self.horizontalLayout_6, 5, 0, 1, 1)

        self.groupBox_12 = QGroupBox(flan_generator_form)
        self.groupBox_12.setObjectName(u"groupBox_12")
        self.gridLayout_13 = QGridLayout(self.groupBox_12)
        self.gridLayout_13.setObjectName(u"gridLayout_13")
        self.eta_cutoff = SliderWidget(self.groupBox_12)
        self.eta_cutoff.setObjectName(u"eta_cutoff")
        self.eta_cutoff.setProperty("min_slider", 1)
        self.eta_cutoff.setProperty("max_slider", 100)
        self.eta_cutoff.setProperty("min_spinbox", 0.000000000000000)
        self.eta_cutoff.setProperty("max_spinbox", 100.000000000000000)
        self.eta_cutoff.setProperty("display_as_float", True)
        self.eta_cutoff.setProperty("slider_step", 1)
        self.eta_cutoff.setProperty("slider_page", 10)
        self.eta_cutoff.setProperty("spinbox_step", 0.010000000000000)
        self.eta_cutoff.setProperty("spinbox_page", 0.100000000000000)

        self.gridLayout_13.addWidget(self.eta_cutoff, 0, 0, 1, 1)


        self.gridLayout.addWidget(self.groupBox_12, 6, 0, 1, 1)


        self.retranslateUi(flan_generator_form)
        self.do_sample.toggled.connect(flan_generator_form.do_sample_toggled)
        self.early_stopping.toggled.connect(flan_generator_form.early_stopping_toggled)
        self.random_seed.toggled.connect(flan_generator_form.random_seed_toggled)
        self.seed.textEdited.connect(flan_generator_form.seed_changed)
        self.model_version.currentTextChanged.connect(flan_generator_form.model_changed)

        self.model_version.setCurrentIndex(-1)


        QMetaObject.connectSlotsByName(flan_generator_form)
    # setupUi

    def retranslateUi(self, flan_generator_form):
        flan_generator_form.setWindowTitle(QCoreApplication.translate("flan_generator_form", u"Form", None))
        self.groupBox.setTitle(QCoreApplication.translate("flan_generator_form", u"Top P", None))
        self.top_p.setProperty("property", QCoreApplication.translate("flan_generator_form", u"top_p", None))
        self.groupBox_2.setTitle(QCoreApplication.translate("flan_generator_form", u"Max length", None))
        self.max_length.setProperty("property", QCoreApplication.translate("flan_generator_form", u"max_length", None))
        self.random_seed.setText(QCoreApplication.translate("flan_generator_form", u"Random seed", None))
        self.groupBox_11.setTitle(QCoreApplication.translate("flan_generator_form", u"Seed", None))
        self.groupBox_5.setTitle(QCoreApplication.translate("flan_generator_form", u"Length penalty", None))
        self.length_penalty.setProperty("property", QCoreApplication.translate("flan_generator_form", u"length_penalty", None))
        self.groupBox_6.setTitle(QCoreApplication.translate("flan_generator_form", u"Num beams", None))
        self.num_beams.setProperty("property", QCoreApplication.translate("flan_generator_form", u"num_beams", None))
        self.do_sample.setText(QCoreApplication.translate("flan_generator_form", u"Do sample", None))
        self.early_stopping.setText(QCoreApplication.translate("flan_generator_form", u"Early stopping", None))
        self.groupBox_3.setTitle(QCoreApplication.translate("flan_generator_form", u"Repetition penalty", None))
        self.repetition_penalty.setProperty("property", QCoreApplication.translate("flan_generator_form", u"repetition_penalty", None))
        self.groupBox_4.setTitle(QCoreApplication.translate("flan_generator_form", u"Min length", None))
        self.min_length.setProperty("property", QCoreApplication.translate("flan_generator_form", u"min_length", None))
        self.groupBox_7.setTitle(QCoreApplication.translate("flan_generator_form", u"No repeat ngram size", None))
        self.ngram_size.setProperty("property", QCoreApplication.translate("flan_generator_form", u"ngram", None))
        self.groupBox_8.setTitle(QCoreApplication.translate("flan_generator_form", u"Temperature", None))
        self.temperature.setProperty("property", QCoreApplication.translate("flan_generator_form", u"temperature", None))
        self.groupBox_9.setTitle(QCoreApplication.translate("flan_generator_form", u"Sequences to generate", None))
        self.sequences.setProperty("property", QCoreApplication.translate("flan_generator_form", u"sequences", None))
        self.groupBox_10.setTitle(QCoreApplication.translate("flan_generator_form", u"Top k", None))
        self.top_k.setProperty("property", QCoreApplication.translate("flan_generator_form", u"top_k", None))
        self.groupBox_12.setTitle(QCoreApplication.translate("flan_generator_form", u"ETA Cutoff", None))
        self.eta_cutoff.setProperty("property", QCoreApplication.translate("flan_generator_form", u"eta_cutoff", None))
    # retranslateUi





from airunner.data.models import AIModels, GeneratorSettings
from airunner.enums import SignalCode, GeneratorSection, ImageGenerator
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.stablediffusion.templates.stable_diffusion_settings_ui import Ui_stable_diffusion_settings_widget
from airunner.windows.main.pipeline_mixin import PipelineMixin
from airunner.utils.create_worker import create_worker
from airunner.workers import ModelScannerWorker
from airunner.settings import AIRUNNER_ART_ENABLED


class StableDiffusionSettingsWidget(
    BaseWidget,
    PipelineMixin
):
    widget_class_ = Ui_stable_diffusion_settings_widget

    def __init__(self, *args, **kwargs):
        self.signal_handlers = {
            SignalCode.AI_MODELS_CREATE_SIGNAL: self.on_models_changed_signal,
            SignalCode.APPLICATION_MAIN_WINDOW_LOADED_SIGNAL: self.update_form,
        }
        super().__init__(*args, **kwargs)
        PipelineMixin.__init__(self)
        self.model_scanner_worker = create_worker(ModelScannerWorker)

    def showEvent(self, event):
        super().showEvent(event)
        self.update_form()

    def update_form(self):
        steps = self.generator_settings.steps
        scale = self.generator_settings.scale

        current_steps = self.get_form_element("steps_widget").property("current_value")
        current_scale = self.get_form_element("scale_widget").property("current_value")

        if steps != current_steps:
            self.get_form_element("steps_widget").setProperty("current_value", steps)

        if scale != current_scale:
            self.get_form_element("scale_widget").setProperty("current_value", scale)

        try:
            self.ui.seed_widget.setProperty("generator_section", self.generator_settings.pipeline_action)
            self.ui.seed_widget.setProperty("generator_name", ImageGenerator.STABLEDIFFUSION.value)
            self.ui.ddim_eta_slider_widget.hide()
            self.ui.frames_slider_widget.hide()
        except RuntimeError as e:
            if AIRUNNER_ART_ENABLED:
                self.logger.error(f"Error updating form: {e}")

        self.model_scanner_worker.add_to_queue("scan_for_models")

        try:
            self.ui.use_compel.setChecked(self.generator_settings.use_compel)
        except RuntimeError as e:
            if AIRUNNER_ART_ENABLED:
                self.logger.error(f"Error updating compel: {e}")

    def toggled_use_compel(self, val):
        self.update_generator_settings("use_compel", val)

    def handle_model_changed(self, _model_name):
        index = self.ui.model.currentIndex()
        model_id = self.ui.model.itemData(index)
        self.update_generator_settings("model", model_id)
        if self.application_settings.sd_enabled:
            self.emit_signal(SignalCode.SD_LOAD_SIGNAL, {
                "do_reload": True
            })

    def handle_scheduler_changed(self, name):
        self.update_generator_settings("scheduler", name)
        self.emit_signal(SignalCode.CHANGE_SCHEDULER_SIGNAL, {"scheduler": name})

    def handle_pipeline_changed(self, val):
        if val == f"{GeneratorSection.TXT2IMG.value} / {GeneratorSection.IMG2IMG.value}":
            val = GeneratorSection.TXT2IMG.value
        elif val == f"{GeneratorSection.INPAINT.value} / {GeneratorSection.OUTPAINT.value}":
            val = GeneratorSection.INPAINT.value
        
        generator_settings = GeneratorSettings.objects.first()
        do_reload = False
        if val == GeneratorSection.TXT2IMG.value:
            model = AIModels.objects.filter_first(
                AIModels.id == generator_settings.model
            )
            if model.pipeline_action == GeneratorSection.INPAINT.value:
                model = AIModels.objects.filter_first(
                    AIModels.version == generator_settings.version,
                    AIModels.pipeline_action == val,
                    AIModels.enabled is True,
                    AIModels.is_default is False
                )
                if model is not None:
                    generator_settings.model = model.id
                else:
                    generator_settings.model = None
                do_reload = True
        generator_settings.pipeline_action = val
        generator_settings.save()
        
        self.load_versions()
        self.load_models()
        if do_reload:
            if self.application_settings.sd_enabled:
                self.emit_signal(SignalCode.SD_LOAD_SIGNAL, {
                    "do_reload": True
                })

    def handle_version_changed(self, val):
        self.update_generator_settings("version", val)
        self.emit_signal(SignalCode.WIDGET_ELEMENT_CHANGED, {
            "element": "sd_version",
            "version": val
        })
        
        generator_settings = GeneratorSettings.objects.first()
        model = AIModels.objects.filter_first(
            AIModels.version == val,
            AIModels.pipeline_action == generator_settings.pipeline_action,
            AIModels.enabled.is_(True),
            AIModels.is_default.is_(False)
        )
        generator_settings.version = val
        generator_settings.model = model.id
        generator_settings.save()
        
        self.load_models()
        if self.application_settings.sd_enabled:
            self.emit_signal(SignalCode.SD_LOAD_SIGNAL, {
                "do_reload": True
            })

    def _load_pipelines(self):
        self.ui.pipeline.blockSignals(True)
        self.ui.pipeline.clear()
        pipeline_names = [
            f"{GeneratorSection.TXT2IMG.value} / {GeneratorSection.IMG2IMG.value}",
            f"{GeneratorSection.INPAINT.value} / {GeneratorSection.OUTPAINT.value}"
        ]
        self.ui.pipeline.addItems(pipeline_names)
        current_pipeline = self.generator_settings.pipeline_action
        if current_pipeline != "":
            if current_pipeline == GeneratorSection.TXT2IMG.value:
                current_pipeline = f"{GeneratorSection.TXT2IMG.value} / {GeneratorSection.IMG2IMG.value}"
            elif current_pipeline == GeneratorSection.INPAINT.value:
                current_pipeline = f"{GeneratorSection.INPAINT.value} / {GeneratorSection.OUTPAINT.value}"
            self.ui.pipeline.setCurrentText(current_pipeline)
        self.ui.pipeline.blockSignals(False)
    
    def load_versions(self):
        self.ui.version.blockSignals(True)
        self.ui.version.clear()
        pipelines = self.get_pipelines(category=ImageGenerator.STABLEDIFFUSION.value)
        version_names = set([pipeline["version"] for pipeline in pipelines])
        self.ui.version.addItems(version_names)
        current_version = self.generator_settings.version
        if current_version != "":
            self.ui.version.setCurrentText(current_version)
        self.ui.version.blockSignals(False)

    def on_models_changed_signal(self):
        try:
            self._load_pipelines()
            self.load_versions()
            self.load_models()
            self.load_schedulers_dropdown()
        except RuntimeError as e:
            if AIRUNNER_ART_ENABLED:
                self.logger.error(f"Error loading models: {e}")

    def clear_models(self):
        self.ui.model.clear()

    def load_models(self):
        self.logger.info("Loading models")
        self.ui.model.blockSignals(True)
        self.clear_models()

        image_generator = ImageGenerator.STABLEDIFFUSION.value
        generator_settings = GeneratorSettings.objects.first()
        pipeline = generator_settings.pipeline_action
        version = generator_settings.version
        pipeline_actions = [GeneratorSection.TXT2IMG.value]
        
        if pipeline == GeneratorSection.INPAINT.value:
            pipeline_actions.append(GeneratorSection.INPAINT.value)
        
        models = AIModels.objects.filter(
            AIModels.category == image_generator,
            AIModels.pipeline_action.in_(pipeline_actions),
            AIModels.version == version,
            AIModels.enabled.is_(True),
            AIModels.is_default.is_(False)
        )

        model_id = generator_settings.model
        if model_id is None and len(models) > 0:
            current_model = models[0]
            generator_settings.model = current_model.id
            generator_settings.save()
        
        for model in models:
            self.ui.model.addItem(model.name, model.id)
        
        if model_id:
            index = self.ui.model.findData(model_id)
            if index != -1:
                self.ui.model.setCurrentIndex(index)
        self.ui.model.blockSignals(False)

    def load_schedulers_dropdown(self):
        self.ui.scheduler.blockSignals(True)
        scheduler_names = [s.display_name for s in self.schedulers]
        self.ui.scheduler.clear()
        self.ui.scheduler.addItems(scheduler_names)

        current_scheduler = self.generator_settings.scheduler
        if current_scheduler != "":
            self.ui.scheduler.setCurrentText(current_scheduler)
        else:
            self.generator_settings.scheduler = self.ui.scheduler.currentText()
        self.update_generator_settings("scheduler", self.generator_settings.scheduler)
        self.ui.scheduler.blockSignals(False)



from airunner.handlers.llm.agent.tools.chat_engine_tool import ChatEngineTool
from airunner.handlers.llm.agent.tools.rag_engine_tool import RAGEngineTool
from airunner.handlers.llm.agent.tools.react_agent_tool import ReActAgentTool


__all__ = [
    "ChatEngineTool",
    "RAGEngineTool",
    "ReActAgentTool",
]



from airunner.handlers.llm.agent.agents.base import BaseAgent
from airunner.handlers.llm.agent.agents.groq import GroqAgent
from airunner.handlers.llm.agent.agents.local import LocalAgent
from airunner.handlers.llm.agent.agents.openai import OpenAI
from airunner.handlers.llm.agent.agents.openrouter import OpenRouterQObject


__all__ = [
    "BaseAgent",
    "GroqAgent",
    "LocalAgent",
    "OpenAI",
    "OpenRouterQObject",
]




from PySide6.QtWidgets import QLineEdit

from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.civitai_preferences.templates.civitai_preferences_widget_ui import Ui_civitai_preferences


class CivitAIPreferencesWidget(BaseWidget):
    widget_class_ = Ui_civitai_preferences

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.ui.api_key.blockSignals(True)
        self.ui.api_key.setEchoMode(QLineEdit.EchoMode.Password)  # Set echo mode to Password
        self.ui.api_key.setText(self.application_settings.civit_ai_api_key)
        self.ui.api_key.blockSignals(False)

    def on_text_changed(self, text):
        self.update_application_settings('civitai_api_key', text)



from typing import Type, Optional

from llama_index.core.llms.llm import LLM
from llama_index.llms.openrouter import OpenRouter
from llama_index.core.chat_engine.types import AgentChatResponse

from airunner.enums import LLMActionType
from airunner.handlers.llm.llm_request import OpenrouterMistralRequest
from airunner.handlers.llm.agent.agents.local import LocalAgent


class OpenRouterQObject(
    LocalAgent
):
    @property
    def llm(self) -> Type[LLM]:
        if not self._llm:
            llm_request = OpenrouterMistralRequest.from_default()
            self._llm = OpenRouter(
                model=self.llm_settings.model,
                api_key=self.llm_settings.openrouter_api_key,
                **llm_request.to_dict()
            )
        return self._llm
    
    def chat(
        self,
        message: str,
        action: LLMActionType = LLMActionType.CHAT,
        system_prompt: Optional[str] = None,
        rag_system_prompt: Optional[str] = None,
        llm_request: Optional[OpenrouterMistralRequest] = None
    ) -> AgentChatResponse:
        llm_request = llm_request or OpenrouterMistralRequest.from_default()
        return super().chat(
            message=message,
            action=action,
            system_prompt=system_prompt,
            rag_system_prompt=rag_system_prompt,
            llm_request=llm_request,
            **llm_request.to_dict()
        )




# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'speecht5_preferences.ui'
##
## Created by: Qt User Interface Compiler version 6.7.0
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QBrush, QColor, QConicalGradient, QCursor,
    QFont, QFontDatabase, QGradient, QIcon,
    QImage, QKeySequence, QLinearGradient, QPainter,
    QPalette, QPixmap, QRadialGradient, QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QGridLayout, QGroupBox,
    QLabel, QSizePolicy, QSpacerItem, QWidget)

from airunner.widgets.slider.slider_widget import SliderWidget

class Ui_speecht5_preferences(object):
    def setupUi(self, speecht5_preferences):
        if not speecht5_preferences.objectName():
            speecht5_preferences.setObjectName(u"speecht5_preferences")
        speecht5_preferences.resize(512, 787)
        self.gridLayout_7 = QGridLayout(speecht5_preferences)
        self.gridLayout_7.setSpacing(0)
        self.gridLayout_7.setObjectName(u"gridLayout_7")
        self.gridLayout_7.setContentsMargins(0, 0, 0, 0)
        self.verticalSpacer = QSpacerItem(20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)

        self.gridLayout_7.addItem(self.verticalSpacer, 10, 0, 1, 1)

        self.preferences_groupbox = QGroupBox(speecht5_preferences)
        self.preferences_groupbox.setObjectName(u"preferences_groupbox")
        self.preferences_groupbox.setCheckable(False)
        self.gridLayout_8 = QGridLayout(self.preferences_groupbox)
        self.gridLayout_8.setObjectName(u"gridLayout_8")
        self.gridLayout_8.setHorizontalSpacing(0)
        self.gridLayout_8.setVerticalSpacing(10)
        self.gridLayout_8.setContentsMargins(10, 10, 10, 10)
        self.label_3 = QLabel(self.preferences_groupbox)
        self.label_3.setObjectName(u"label_3")
        font = QFont()
        font.setPointSize(9)
        self.label_3.setFont(font)

        self.gridLayout_8.addWidget(self.label_3, 0, 0, 1, 1)

        self.pitch = SliderWidget(self.preferences_groupbox)
        self.pitch.setObjectName(u"pitch")
        self.pitch.setProperty("slider_minimum", 1)
        self.pitch.setProperty("slider_maximum", 100)
        self.pitch.setProperty("spinbox_minimum", 0.000000000000000)
        self.pitch.setProperty("spinbox_maximum", 1.000000000000000)
        self.pitch.setProperty("display_as_float", True)
        self.pitch.setProperty("slider_single_step", 1)
        self.pitch.setProperty("slider_page_step", 10)
        self.pitch.setProperty("spinbox_single_step", 0.010000000000000)
        self.pitch.setProperty("spinbox_page_step", 0.100000000000000)

        self.gridLayout_8.addWidget(self.pitch, 2, 0, 1, 1)

        self.groupBox = QGroupBox(self.preferences_groupbox)
        self.groupBox.setObjectName(u"groupBox")
        self.gridLayout = QGridLayout(self.groupBox)
        self.gridLayout.setObjectName(u"gridLayout")
        self.voice = QComboBox(self.groupBox)
        self.voice.addItem("")
        self.voice.addItem("")
        self.voice.addItem("")
        self.voice.addItem("")
        self.voice.addItem("")
        self.voice.addItem("")
        self.voice.addItem("")
        self.voice.setObjectName(u"voice")

        self.gridLayout.addWidget(self.voice, 0, 0, 1, 1)


        self.gridLayout_8.addWidget(self.groupBox, 1, 0, 1, 1)


        self.gridLayout_7.addWidget(self.preferences_groupbox, 0, 0, 1, 1)


        self.retranslateUi(speecht5_preferences)
        self.voice.currentTextChanged.connect(speecht5_preferences.voice_changed)

        QMetaObject.connectSlotsByName(speecht5_preferences)
    # setupUi

    def retranslateUi(self, speecht5_preferences):
        speecht5_preferences.setWindowTitle(QCoreApplication.translate("speecht5_preferences", u"Form", None))
        self.preferences_groupbox.setTitle(QCoreApplication.translate("speecht5_preferences", u"SpeechT5", None))
        self.label_3.setText(QCoreApplication.translate("speecht5_preferences", u"More realistic. Slower. Uses VRAM", None))
        self.pitch.setProperty("settings_property", QCoreApplication.translate("speecht5_preferences", u"speech_t5_settings.pitch", None))
        self.pitch.setProperty("label_text", QCoreApplication.translate("speecht5_preferences", u"Pitch", None))
        self.groupBox.setTitle(QCoreApplication.translate("speecht5_preferences", u"Voice Style", None))
        self.voice.setItemText(0, QCoreApplication.translate("speecht5_preferences", u"US Male", None))
        self.voice.setItemText(1, QCoreApplication.translate("speecht5_preferences", u"US Male 2", None))
        self.voice.setItemText(2, QCoreApplication.translate("speecht5_preferences", u"US Female", None))
        self.voice.setItemText(3, QCoreApplication.translate("speecht5_preferences", u"US Female 2", None))
        self.voice.setItemText(4, QCoreApplication.translate("speecht5_preferences", u"Canadian Male", None))
        self.voice.setItemText(5, QCoreApplication.translate("speecht5_preferences", u"Scottish Male", None))
        self.voice.setItemText(6, QCoreApplication.translate("speecht5_preferences", u"Indian Male", None))

    # retranslateUi




from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.orm import relationship

from airunner.data.models.base import BaseModel


class TargetFiles(BaseModel):
    __tablename__ = 'target_files'
    id = Column(Integer, primary_key=True, autoincrement=True)
    chatbot_id = Column(Integer, ForeignKey('chatbots.id'))
    file_path = Column(String)

    chatbot = relationship("Chatbot", back_populates="target_files")



from typing import Optional

from PIL.ImageQt import QImage

from PySide6.QtCore import QRect, QPoint
from PySide6.QtGui import QBrush, QColor, QPen, QPixmap, QPainter, Qt
from PySide6.QtWidgets import QGraphicsItem

from airunner.enums import SignalCode, CanvasToolName
from airunner.widgets.canvas.draggables.draggable_pixmap import DraggablePixmap


class ActiveGridArea(DraggablePixmap):
    def __init__(self):
        self.image = None
        self._current_width = 0
        self._current_height = 0
        self._render_border: bool = False
        self._line_width: int = 1
        self._do_draw = True
        self._do_render_fill = False
        self._active_grid_settings_enabled = False
        self._draggable_rect: Optional[QRect] = None
        self._border_pen: Optional[QPen] = None
        self._outer_border_pen: Optional[QPen] = None
        self._border_color: Optional[QColor] = None
        self._border_brush: Optional[QBrush] = None

        super().__init__(QPixmap())
        self.render_fill()

        painter = self.draw_border()
        super().paint(painter, None, None)

        self.snap_to_grid(
            x=min(self.rect.x(), self.rect.x() + self.rect.width()),
            y=min(self.rect.y(), self.rect.y() + self.rect.height()),
            save=True
        )
        self.setFlag(
            QGraphicsItem.GraphicsItemFlag.ItemIsMovable,
            True
        )
        self.register(
            SignalCode.APPLICATION_SETTINGS_CHANGED_SIGNAL,
            self.render_fill
        )

    def update_position(self, x: int, y: int, save: bool = True):
        self.setPos(QPoint(x, y))
        if save:
            self.update_active_grid_settings("pos_x", x)
            self.update_active_grid_settings("pos_y", y)

    @property
    def rect(self):
        return QRect(
            self.active_grid_settings.pos_x,
            self.active_grid_settings.pos_y,
            self.application_settings.working_width,
            self.application_settings.working_height
        )

    def render_fill(self):
        self._current_width = self.rect.width()
        self._current_height = self.rect.height()
        self._do_render_fill = self.active_grid_settings.render_fill
        self._active_grid_settings_enabled = self.active_grid_settings.enabled

        width = abs(self.rect.width())
        height = abs(self.rect.height())

        if not self.image:
            self.image = QImage(width, height, QImage.Format.Format_ARGB32)
        else:
            self.image = self.image.scaled(width, height)

        fill_color = self.get_fill_color() if self._do_render_fill else QColor(0, 0, 0, 1)
        self.image.fill(fill_color)
        pixmap = QPixmap.fromImage(self.image)
        self.setPixmap(pixmap)

    def get_fill_color(self) -> QColor:
        render_fill = self.active_grid_settings.render_fill
        if render_fill:
            fill_color = self.active_grid_settings.fill_color
            fill_color = QColor(fill_color)
            fill_opacity = self.active_grid_settings.fill_opacity
            fill_opacity = max(1, fill_opacity)
            fill_color.setAlpha(fill_opacity)
        else:
            fill_color = QColor(0, 0, 0, 1)
        return fill_color

    def toggle_render_fill(self, _render_fill):
        self.update_selection_fill()

    def change_fill_opacity(self, _value):
        self.update_selection_fill()

    def update_selection_fill(self):
        self.pixmap.fill(self.get_fill_color())

    def draw_border(self, painter: QPainter = None) -> QPainter:
        if painter is None:
            painter = QPainter(self.pixmap)

        if self.active_grid_settings.enabled:
            render_border = self.active_grid_settings.render_border
            line_width = self.grid_settings.line_width

            self._draggable_rect = QRect(
                0,
                0,
                abs(self.rect.width()),
                abs(self.rect.height())
            )
            border_color = QColor(self.active_grid_settings.border_color)
            border_color.setAlpha(self.active_grid_settings.border_opacity)
            self._border_pen = QPen(
                border_color,
                line_width
            )
            self._outer_border_pen = QPen(
                border_color,
                line_width + 1
            )
            self._border_color = QColor(0, 0, 0, 0)
            self._border_brush = QBrush(self._border_color)

            if render_border:
                painter.setPen(self._border_pen)
                painter.setBrush(self._border_brush)
                painter.drawRect(self._draggable_rect)
                painter.setPen(self._outer_border_pen)
                painter.drawRect(self._draggable_rect)
        return painter

    def paint(self, painter: QPainter, option, widget=None):
        painter = self.draw_border(painter)
        super().paint(painter, option, widget)

    def toggle_render_border(self, value):
        pass

    def change_border_opacity(self, value):
        pass

    mouse_press_pos = None

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        if event.button() == Qt.MouseButton.LeftButton:
            self.mouse_press_pos = event.pos()

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)
        if self.mouse_press_pos and self.current_tool is CanvasToolName.ACTIVE_GRID_AREA and (
            self.mouse_press_pos.x() != event.pos().x() or
            self.mouse_press_pos.y() != event.pos().y()
        ):
            self.emit_signal(SignalCode.ACTIVE_GRID_AREA_MOVED_SIGNAL)
            self.emit_signal(SignalCode.GENERATE_MASK)
        self.mouse_press_pos = None

    def mouseMoveEvent(self, event):
        if self.current_tool not in [
            CanvasToolName.ACTIVE_GRID_AREA,
        ]:
            return
        super().mouseMoveEvent(event)



from typing import Optional

from airunner.app import App
from airunner.handlers.llm import (
    LLMRequest, 
    LLMResponse
)
from airunner.handlers.stablediffusion.image_request import (
    ImageRequest
)
from airunner.enums import (
    SignalCode, 
    LLMActionType
)


class API(App):
    def send_llm_request(
        self, 
        prompt: str, 
        llm_request: Optional[LLMRequest] = None,
        action: LLMActionType = LLMActionType.CHAT,
        do_tts_reply: bool = True,
    ):
        """
        Send a request to the LLM with the given prompt and action.
        
        :param prompt: The prompt to send to the LLM.
        :param llm_request: Optional LLMRequest object.
        :param action: The action type for the request.
        :param do_tts_reply: Whether to do text-to-speech reply.
        :return: None
        """
        llm_request = llm_request or LLMRequest.from_default()
        llm_request.do_tts_reply = do_tts_reply

        self.emit_signal(
            SignalCode.LLM_TEXT_GENERATE_REQUEST_SIGNAL,
            {
                "llm_request": True,
                "request_data": {
                    "action": action,
                    "prompt": prompt,
                    "llm_request": llm_request,
                    "do_tts_reply": do_tts_reply,
                }
            }
        )
    
    def send_tts_request(self, response: LLMResponse):
        """
        Send a TTS request with the given response."
        
        :param response: The LLMResponse object.
        :return: None
        """
        self.emit_signal(SignalCode.LLM_TEXT_STREAMED_SIGNAL, {
            "response": response
        })
    
    def send_image_request(
        self, 
        image_request: Optional[ImageRequest] = None
    ):
        """"
        Send a request to the image generator with the given request.
        :param sd_request: Optional ImageRequest object.
        :return: None
        """
        image_request = image_request or ImageRequest()
        self.emit_signal(SignalCode.DO_GENERATE_SIGNAL, {
            "sd_request": image_request
        })


import os

from PySide6.QtGui import QMovie
from PySide6.QtCore import QSize

from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.llm.templates.loading_ui import Ui_loading_message


class LoadingWidget(BaseWidget):
    widget_class_ = Ui_loading_message

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        here = os.path.dirname(os.path.abspath(__file__))
        movie = QMovie(os.path.join(here, "../../icons/dark/Spinner-1s-200px.gif"))
        movie.setScaledSize(QSize(64, 64))  # Resize the GIF
        self.ui.label.setMovie(movie)  # Set the QMovie object to the label
        movie.start()  # Start the animation

    def set_size(self, spinner_size: QSize, label_size: QSize):
        self.ui.label.movie().setScaledSize(spinner_size)
        self.ui.label.setFixedSize(label_size)
        self.ui.label.update()
        self.ui.label.repaint()



from sqlalchemy import Column, Integer, String, Boolean
from sqlalchemy.orm import relationship

from airunner.data.models.base import BaseModel


class AIModels(BaseModel):
    __tablename__ = 'aimodels'
    id = Column(Integer, primary_key=True, autoincrement=True)
    name = Column(String, nullable=False, default="Model")
    path = Column(String, nullable=False, default="")
    branch = Column(String, nullable=False, default="")
    version = Column(String, nullable=False, default="")
    category = Column(String, nullable=False, default="")
    pipeline_action = Column(String, nullable=False, default="")
    enabled = Column(Boolean, nullable=False, default=True)
    model_type = Column(String, nullable=False, default="")
    is_default = Column(Boolean, nullable=False, default=False)
    generator_settings = relationship("GeneratorSettings", back_populates="aimodel")




from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.image.templates.folder_widget_ui import Ui_folder_widget


class FolderWidget(BaseWidget):
    widget_class_ = Ui_folder_widget
    path = None
    callback = None

    def set_path(self, path):
        self.path = path
        self.ui.label.setText(path)

    def folder_clicked(self):
        self.callback(self.path)



from airunner.handlers.llm.huggingface_llm import HuggingFaceLLM
from airunner.handlers.llm.llm_handler import LLMHandler
from airunner.handlers.llm.llm_request import LLMRequest
from airunner.handlers.llm.llm_response import LLMResponse
from airunner.handlers.llm.llm_settings import LLMSettings
from airunner.handlers.llm.training_mixin import TrainingMixin
from airunner.handlers.llm.llm_request import OpenrouterMistralRequest


__all__ = [
    "HuggingFaceLLM",
    "LLMHandler",
    "LLMRequest",
    "LLMResponse",
    "LLMSettings",
    "TrainingMixin",
    "OpenrouterMistralRequest",
]



from airunner.enums import SignalCode
from airunner.widgets.base_widget import BaseWidget
from airunner.widgets.paths.templates.paths_ui import Ui_paths_form


class PathsWidget(BaseWidget):
    widget_class_ = Ui_paths_form

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def action_button_clicked_reset(self):
        self.emit_signal(SignalCode.APPLICATION_RESET_PATHS_SIGNAL)



from sqlalchemy import Column, Integer, Boolean, String

from airunner.data.models.base import BaseModel


class TTSSettings(BaseModel):
    __tablename__ = 'tts_settings'
    id = Column(Integer, primary_key=True, autoincrement=True)
    tts_model = Column(String, default="SpeechT5")
    use_cuda = Column(Boolean, default=True)
    use_sentence_chunks = Column(Boolean, default=True)
    use_word_chunks = Column(Boolean, default=False)
    cuda_index = Column(Integer, default=0)
    word_chunks = Column(Integer, default=1)
    sentence_chunks = Column(Integer, default=1)
    model = Column(String, default="SpeechT5")
